"""
TEST_MATRICES.

Tests for the Integer, IntegerPython, IntegerRing, and IntegerRingPython
classes.
"""

import pytest
import timeit
import numpy as np
from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.rationals import QQ
from jacamar.rings.reals import RR, RR_py
from jacamar.rings.complexes import CC
from jacamar.rings.polynomials import PolynomialRing
from jacamar.matrices.matrices import Matrix, _MatrixGenericData, generate, random


class TestMatrix:
    """Tests for the Matrix class."""

    m = Matrix(base_ring=ZZ, entries=[[1, 2], [3, 4]])
    q = Matrix(base_ring=ZZ, entries=[[1, 0], [3, 0]])
    n = Matrix(base_ring=ZZ, entries=[[1, 2, 3], [4, 5, 6]])
    thin = Matrix(base_ring=ZZ, nrows=3, ncols=0)
    flat = Matrix(base_ring=ZZ, nrows=0, ncols=2)

    # Generics
    m_py = Matrix(base_ring=ZZ_py, entries=[[1, 2], [3, 4]])
    q_py = Matrix(base_ring=ZZ_py, entries=[[1, 0], [3, 0]])
    n_py = Matrix(base_ring=ZZ_py, entries=[[1, 2, 3], [4, 5, 6]])
    thin_py = Matrix(base_ring=ZZ_py, nrows=3, ncols=0)
    flat_py = Matrix(base_ring=ZZ_py, nrows=0, ncols=2)

    def test_construction_from_jacamar(self):
        """Tests for construction from JACAMAR elements."""
        assert self.m == Matrix(base_ring=ZZ, entries=[[ZZ(1), ZZ(2)], [ZZ(3), ZZ(4)]])
        assert self.m._is_generic == False

    def test_classes(self):
        """Tests classes."""
        assert isinstance(self.m, Matrix)

    def test_equality_among_the_classes(self):
        pass

    def test_sizes(self):
        """Tests nrows and ncols."""
        assert self.m.nrows == 2
        assert self.m.ncols == 2
        assert self.n.nrows == 2
        assert self.n.ncols == 3
        assert self.thin.nrows == 3
        assert self.thin.ncols == 0
        assert self.flat.nrows == 0
        assert self.flat.ncols == 2

        assert self.m_py.nrows == 2
        assert self.m_py.ncols == 2
        assert self.n_py.nrows == 2
        assert self.n_py.ncols == 3
        assert self.thin_py.nrows == 3
        assert self.thin_py.ncols == 0
        assert self.flat_py.nrows == 0
        assert self.flat_py.ncols == 2

    def test_add(self):
        """Tests __add__."""
        assert self.m + self.m == Matrix(base_ring=ZZ, entries=[[2, 4], [6, 8]])
        assert self.m * ZZ(2)
        assert self.m_py + self.m_py == Matrix(
            base_ring=ZZ_py, entries=[[2, 4], [6, 8]]
        )
        with pytest.raises(ValueError):
            self.m + self.n
        with pytest.raises(ValueError):
            self.m_py + self.n_py
        with pytest.raises(TypeError):
            self.m + self.m_py
        with pytest.raises(TypeError):
            self.m_py + self.m

    def test_mul(self):
        """Tests __mul__."""
        # x = self.m * self.n
        # x_py = self.m_py * self.n_py
        # assert x == Matrix(base_ring=ZZ, entries=[[9, 12, 15], [19, 26, 33]])
        # assert x_py == Matrix(base_ring=ZZ_py, entries=[[9, 12, 15], [19, 26, 33]])
        s1 = 0
        s2 = 0
        s3 = 17
        a = Matrix(base_ring=RR, entries=np.ones((s1, s1), dtype=int).tolist())
        b = Matrix(base_ring=RR, entries=np.ones((s2, s2), dtype=int).tolist())
        c = Matrix(base_ring=RR, entries=np.ones((s3, s3), dtype=int).tolist())
        assert Matrix(
            base_ring=RR, entries=(-1 * np.ones((s1, s1), dtype=int)).tolist()
        ) == a * RR(-1)
        assert Matrix(
            base_ring=ZZ, entries=(-1 * np.ones((s1, s1), dtype=int)).tolist()
        ) == a * ZZ(-1)
        assert a * a == Matrix(
            base_ring=RR, entries=(s1 * np.ones((s1, s1), dtype=int)).tolist()
        )
        assert b * b == Matrix(
            base_ring=RR, entries=(s2 * np.ones((s2, s2), dtype=int)).tolist()
        )
        assert c * c == Matrix(
            base_ring=RR, entries=(s3 * np.ones((s3, s3), dtype=int)).tolist()
        )

    def test_empty_mul(self):
        """Tests multiplication with empty matrices of various sizes."""

        y = self.flat * self.m
        y_py = self.flat_py * self.m_py

        assert y.nrows == 0
        assert y.ncols == 2
        assert y_py.nrows == 0
        assert y_py.ncols == 2

        assert y == Matrix(base_ring=ZZ, nrows=0, ncols=2)
        assert y_py == Matrix(base_ring=ZZ_py, nrows=0, ncols=2)

        z = self.n * self.thin
        z_py = self.n_py * self.thin_py

        assert z.nrows == 2
        assert z.ncols == 0
        assert z_py.nrows == 2
        assert z_py.ncols == 0

        assert z == Matrix(base_ring=ZZ, nrows=2, ncols=0)
        assert z_py == Matrix(base_ring=ZZ_py, nrows=2, ncols=0)

        w = Matrix(base_ring=ZZ, nrows=0, ncols=0)
        w_py = Matrix(base_ring=ZZ_py, nrows=0, ncols=0)

        assert w * w == w
        assert w_py * w_py == w_py

        assert w.nrows == 0
        assert w.ncols == 0
        assert w_py.nrows == 0
        assert w_py.ncols == 0

    def test_sub(self):
        """Tests __sub__."""
        assert self.m - self.q == Matrix(base_ring=ZZ, entries=[[0, 2], [0, 4]])
        assert self.m_py - self.q_py == Matrix(
            base_ring=ZZ_py, entries=[[0, 2], [0, 4]]
        )
        with pytest.raises(ValueError):
            self.m - self.n
        with pytest.raises(ValueError):
            self.m_py - self.n_py
        with pytest.raises(TypeError):
            self.m - self.m_py
        with pytest.raises(TypeError):
            self.m_py - self.m

    def test_rational(self):
        """Tests QQ with matrices."""
        a = Matrix(base_ring=QQ, entries=[[QQ(1, 2), QQ(3, 5)], [QQ(46, 3), QQ(23, 2)]])
        b = Matrix(
            base_ring=QQ, entries=[[QQ(189, 20), QQ(36, 5)], [QQ(184), QQ(2829, 20)]]
        )
        c = Matrix(base_ring=QQ, entries=[[QQ(1), QQ(6, 5)], [QQ(92, 3), QQ(23)]])
        assert a * a == b
        assert a + a == c

    def test_getitem(self):
        """Tests __getitem__."""
        a = Matrix(base_ring=RR, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b = Matrix(base_ring=RR, entries=[[5]])
        a_py = Matrix(base_ring=RR_py, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b_py = Matrix(base_ring=RR_py, entries=[[5]])
        assert a[0, 2] == RR(3)
        assert b[0, 0] == RR(5)
        assert a_py[1, 2] == RR_py(6)
        assert b_py[0, 0] == RR_py(5)

    def test_submatrix(self):
        """Tests __getitem__ with slices."""
        a = Matrix(base_ring=RR, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        c = Matrix(base_ring=RR, entries=[[3], [6], [9]])
        d = Matrix(base_ring=RR, entries=[[4, 5]])
        a_py = Matrix(base_ring=RR_py, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        c_py = Matrix(base_ring=RR_py, entries=[[3], [6], [9]])
        d_py = Matrix(base_ring=RR_py, entries=[[4, 5]])
        assert a_py[:, :] == a_py
        assert a_py[:, 2] == c_py
        assert a_py[1, 0:2] == d_py
        assert a[:, :] == a
        assert a[:, 2] == c
        assert a[1, 0:2] == d

    def test_size(self):
        """Tests .size()"""
        a = Matrix(base_ring=RR, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b = Matrix(base_ring=RR_py, entries=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        assert a.size() == (3, 3)
        assert b.size() == (3, 3)

    def test_det(self):
        """Tests flint determiant method."""
        a = Matrix(base_ring=RR, entries=[[1, 2], [2, 1]])
        b = Matrix(base_ring=RR_py, entries=[[1, 2], [2, 1]])

        assert b.det()
        assert a.det() == RR(-3)

    def test_transpose(self):
        """Tests flint transpose method."""
        a = Matrix(base_ring=RR, entries=[[1, 2], [3, 4]])
        b = Matrix(base_ring=RR, entries=[[1, 3], [2, 4]])
        a_py = Matrix(base_ring=RR_py, entries=[[1, 2], [3, 4]])
        b_py = Matrix(base_ring=RR_py, entries=[[1, 3], [2, 4]])
        assert a_py.transpose() == b_py
        assert a.T() == b

    @pytest.mark.slow
    def test_np_construction(self):
        """Tests numpy construction for RR_py and ZZ_py."""
        a = Matrix(base_ring=ZZ_py, entries=[[1, 2], [3, 4]])
        s = 1000
        r = random(RR_py, 10, s, s)
        z = random(ZZ_py, 10, s, s)

        assert r * r
        assert z * z


class TestGenericMatrices:
    """Tests for generic matrices."""

    z = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x", packed=False).to_generic()
    x0 = z.gens[0]
    x1 = z.gens[1]
    x2 = z.gens[2]
    zreg = x0 + x1 * x2
    r = PolynomialRing(base_ring=RR, ngens=3, prefix="x", packed=False).to_generic()
    y0 = r.gens[0]
    y1 = r.gens[1]
    y2 = r.gens[2]

    s = PolynomialRing(base_ring=ZZ, ngens=3, prefix="x")
    x0 = s.gens[0]
    x1 = s.gens[1]
    x2 = s.gens[2]
    sp = x0 + x1 * x2

    # ZZ tests
    def test_generic_creation(self):
        """Test creation of a generic ZZ matrix of polynomials."""
        f = ZZ(2) * self.x1 * self.x2 + ZZ(9) * self.x0 ** ZZ(4)
        a = Matrix(
            base_ring=self.s,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=self.s, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        b = Matrix(base_ring=self.s, ncols=2, nrows=2, entries=[[f, f], [f, f]])
        assert a.data.entries == b.data.entries
        assert a.data == b.data
        assert a == b
        assert f == ZZ(2) * self.x1 * self.x2 + ZZ(9) * self.x0 ** ZZ(4)
        assert a[1, 1] == f

    def test_generic_mult(self):
        """Tests __mul__ of a generic ZZ matrix."""
        f = self.z({(1, 1, 2, 1): ZZ(2), (0, 4): ZZ(9)})
        s = 16
        a = generate(f, s, s)
        b = generate(ZZ(s) * f * f, s, s)
        assert a * a == b

    def test_generic_add(self):
        """Tests __add__ of a generic ZZ matrix."""
        f = self.z({(1, 1, 2, 1): ZZ(2), (0, 4): ZZ(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a + a == Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f + f, f + f], [f + f, f + f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing,
                ncols=2,
                nrows=2,
                entries=[[f + f, f + f], [f + f, f + f]],
            ),
        )

    def test_generic_sub(self):
        """Tests __sub__ of a generic ZZ matrix."""
        f = self.z({(1, 1, 2, 1): ZZ(2), (0, 4): ZZ(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a + a - a == a

    @pytest.mark.slow
    def test_straussen_mult(self):
        """Tests __mul__ (strassen alogorithm) of a generic ZZ matrix."""

        # f = self.z({(1, 1, 1, 2, 1, 4, 1, 5): RR(2)})
        # q = self.z({(1, 1): RR(2)})
        s = 128
        # mq = random(RR, 3, s, s, 1, PolynomialRing, 4)
        # print(type(mq[:, :]))
        # mq = generate(q, s, s)
        # print(timeit.timeit(lambda: q*q, number=int(s**2.81)))
        # print(timeit.timeit(lambda: q*q, number=int(s**3)))
        # assert mq * mq == generate(RR(s)*q*q, s, s)
        # print(timeit.timeit(lambda: mq*mq, number=1))
        # mq = generate(q, s//2, s//2)

        # print(timeit.timeit(lambda: mq*mq, number=8))
        mq = generate(self.zreg, s, s)

        print(timeit.timeit(lambda: mq * mq, number=1))
        mq = generate(self.sp, s, s)

        print(timeit.timeit(lambda: mq * mq, number=1))

        # assert 1 == 0
        # assert mq * mq == generate(ZZ(s)*q*q, s, s)

    def test_kmb_mult(self):
        """Tests __mul__ (Kauers-Moosbauer alogorithm) of a generic ZZ matrix."""
        f = self.z({(1, 1, 2, 1): ZZ(2), (0, 4): ZZ(9)})
        s = 5
        a = generate(f, s, s)
        b = generate(ZZ(s) * f * f, s, s)
        print(len((a * a).data.entries))
        assert a * a == b

    # RR tests
    def test_generic_creation_RR(self):
        """Test creation of a generic RR matrix of polynomials."""
        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = Matrix(
            base_ring=self.r,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a[1, 1] == f

    def test_generic_mult_RR(self):
        """Tests __mul__ of a generic ZZ matrix."""
        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a * a == Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[RR(2) * f * f, RR(2) * f * f], [RR(2) * f * f, RR(2) * f * f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing,
                ncols=2,
                nrows=2,
                entries=[
                    [RR(2) * f * f, RR(2) * f * f],
                    [RR(2) * f * f, RR(2) * f * f],
                ],
            ),
        )

    def test_generic_add_RR(self):
        """Tests __add__ of a generic ZZ matrix."""
        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a + a == Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f + f, f + f], [f + f, f + f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing,
                ncols=2,
                nrows=2,
                entries=[[f + f, f + f], [f + f, f + f]],
            ),
        )

    def test_straussen_mult_RR(self):
        """Tests __mul__ (strassen alogorithm) of a generic ZZ matrix."""

        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a * a == Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[RR(2) * f * f, RR(2) * f * f], [RR(2) * f * f, RR(2) * f * f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing,
                ncols=2,
                nrows=2,
                entries=[
                    [RR(2) * f * f, RR(2) * f * f],
                    [RR(2) * f * f, RR(2) * f * f],
                ],
            ),
        )

    def test_kmb_mult_RR(self):
        """Tests __mul__ (Kauers-Moosbauer alogorithm) of a generic ZZ matrix."""

        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[f, f], [f, f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing, ncols=2, nrows=2, entries=[[f, f], [f, f]]
            ),
        )
        assert a * a == Matrix(
            base_ring=PolynomialRing,
            ncols=2,
            nrows=2,
            entries=[[RR(2) * f * f, RR(2) * f * f], [RR(2) * f * f, RR(2) * f * f]],
            data=_MatrixGenericData(
                base_ring=PolynomialRing,
                ncols=2,
                nrows=2,
                entries=[
                    [RR(2) * f * f, RR(2) * f * f],
                    [RR(2) * f * f, RR(2) * f * f],
                ],
            ),
        )

    def test_generic_det(self):
        """Tests generic implemnetation of the deternimant method for polynomials."""
        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})
        a = generate(f, 2, 2)
        assert a.det() == f * (f ** ZZ(2) - f ** ZZ(2)) - f * (
            f ** ZZ(2) - f ** ZZ(2)
        ) + f * (f ** ZZ(2) - f ** ZZ(2))

    def test_generate(self):
        """Tests generate fuinction, which creates a matrix of size (r, c) with a specific entry."""
        f = self.r({(1, 1, 2, 1): RR(2), (0, 4): RR(9)})

    def test_random(self):
        """Tests random matrix generation function."""
        f = random(RR, 10, 5, 5, 1, PolynomialRing, 3)
        a = random(RR, 10, 10, 10)
        b = random(RR, 2, 10, 10)
        assert a * b, f



"""
SERIES

This is a module for absolutely capped precision power series in multiple variables with arbitrary
positive integer weights.

To model k[[x_1,...,x_n]] we include it into k[y_1,...,y_n][[T]] x_i maps to y_i*T^w_i
where w_i is the weight of x_i. Under this inclusion, the power series ring is closed under all
arithmetic operations in the ambient ring.

Exposed classes:
 - PowerSeriesRing,
 - PowerSeriesRingHomomorphism, and
 - Series.

AUTHORS:
- Benjamin Antieau (2024): initial version.

"""

import functools
import itertools
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.integers import ZZ
from jacamar.rings.rings import AbstractRing
from jacamar.rings.polynomials import (
    PolynomialRing,
    PolynomialData,
    Polynomial,
    MonomialData,
)
from jacamar.rings.morphisms import AbstractRingMorphism


class SeriesData:
    """
    Data class for elements of weighted power series ring.
    """

    def __init__(self, base_ring, term_list, precision):
        """
        The argument term_list is an ordered list [(N,coefficient)] representing g*T^N
        where N is a Python integer and g is a PolynomialData of degree N.
        """
        self.base_ring = base_ring
        self.term_list = term_list
        self.precision = precision

    def __str__(self):
        return str(self.term_list)

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        """
        Add self to other, taking other.precision as the result's precision
        cap.
        """
        term_dict = dict(other.term_list)
        for deg, coeff in self.term_list:
            if deg < other.precision:
                if deg in term_dict:
                    term_dict[deg] += coeff
                else:
                    term_dict[deg] = coeff
        sorted_degrees = list(term_dict)
        sorted_degrees.sort()
        newterm_list = []
        for deg in sorted_degrees:
            if not term_dict[deg].is_zero():
                newterm_list.append((deg, term_dict[deg]))
        return other.__class__(self.base_ring, newterm_list, other.precision)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        return self.__class__(
            self.base_ring,
            [(deg, -coeff) for deg, coeff in self.term_list],
            self.precision,
        )

    def __eq__(self, other):
        return (self.term_list == other.term_list) & (self.precision == other.precision)

    def __ne__(self, other):
        return not self == other

    def is_unit(self):
        if (
            self.term_list[0][0] != 0
            or not self.term_list[0][1].constant_coefficient().is_unit()
        ):
            return False
        else:
            return True

    def __rmul__(self, other):
        # TODO: this could result in zero terms.
        return self.__class__(
            self.base_ring,
            [(deg, other * coefficient) for deg, coefficient in self.term_list],
            self.precision,
        )

    def __mul__(self, other):
        """
        Multiplication assumes that the input term_lists are sorted by degree.

        TODO: this can be parallelized in an obvious way.
        """
        term_dict = {}
        for deg, coeff in self.term_list:
            for deg1, coeff1 in other.term_list:
                if deg + deg1 < other.precision:
                    if deg + deg1 in term_dict:
                        term_dict[deg + deg1] += coeff * coeff1
                    else:
                        term_dict[deg + deg1] = coeff * coeff1
                else:
                    break
        sorted_degrees = list(term_dict)
        sorted_degrees.sort()
        newterm_list = []
        for deg in sorted_degrees:
            if not term_dict[deg].is_zero():
                newterm_list.append((deg, term_dict[deg]))
        return self.__class__(self.base_ring, newterm_list, other.precision)

    def __pow__(self, n):
        # This is probably not very pythonic. But, it is the only
        # way I can figure out how to get self**n to work. The only other option
        # would be to implement __rpow__(self,n) and apply it n**self,
        # which would make for garbage code.
        # From sage/arith/power.pyx
        apow = self
        while not n & 1:  # While even...
            apow *= apow
            # Bitshift
            n >>= 1

        # Now multiply together the correct factors a^(2^i)
        res = apow
        n >>= 1
        while n:
            apow *= apow
            if n & 1:
                res = apow * res
            n >>= 1
        return res

    def _naive_power(self, n):
        pass

    def is_homogeneous(self):
        """
        Constants and zero count as homogeneous.
        """
        non_zero_terms = 0
        for _, coeff in self.term_list:
            if not coeff.is_zero():
                non_zero_terms += 1
            if non_zero_terms > 1:
                return False
        return True

    def degree(self):
        degree = -1
        for deg, coeff in self.term_list:
            if not coeff.is_zero():
                degree = deg
        return degree


class Series(AbstractRingElement):
    data_class = SeriesData

    def __init__(self, ring, data):
        self.base_ring = ring.base_ring
        AbstractRingElement.__init__(
            self,
            ring,
            data,
        )

    def term_data(self, a=None, b=None):
        """
        An iterator returning the terms of the series, packaged as tuples (m,c)
        where m is a MonomialData and c is an element of the base_ring.

        If a and b are provided, return only those terms of total degree
        between a and b.
        """
        iterators = []
        for n, g in self.data.term_list:
            if a <= n <= b:
                iterators.append(g.term_data())
        return itertools.chain.from_iterable(iterators)

    def constant_coefficient(self):
        if self.data.term_list[0][0] != 0:
            return self.base_ring.zero
        else:
            for m, c in self.data.term_list[0][1].monomial_dictionary.items():
                return self.base_ring(c)

    def is_unit(self):
        if self.data.term_list[0][0] == 0:
            for m, c in self.data.term_list[0][1].monomial_dictionary.items():
                if self.base_ring(c).is_unit():
                    return True
        return False

    def inverse(self):
        newton_approximation_inverse = self.ring(self.constant_coefficient())
        N = 1
        while N < self.ring.precision_cap:
            N = 2 * N
            newton_approximation_inverse = (
                newton_approximation_inverse
                + newton_approximation_inverse
                * (self.ring.one - self * newton_approximation_inverse)
            )
        return newton_approximation_inverse

    def __rmul__(self, other):
        return other.__class__(other.ring, self.data * other.data)

    def __str__(self):
        """
        This will return a string representing the underlying element
        in whatever order the coefficient_dictionary has stored its keys in. In particular
        it is not guaranteed to be in lexicographic ordering or anything else useful.
        """
        if len(self.data.term_list) == 0:
            return f"0 + O(F^{self.ring.precision_cap})"
        else:
            return_str = ""
            for _, coef in self.data.term_list:
                return_str += str(self.ring._polynomial_ring(coef)) + " + "
            return return_str + f"O(F^{self.ring.precision_cap})"

    def __repr__(self):
        return self.__str__()


class PowerSeriesRing(AbstractRing):
    """
    Weighted power series ring.
    """

    element_class = Series

    def __init__(
        self,
        *,
        base_ring,
        ngens,
        prefix,
        precision_cap,
        weights=None,
    ):
        """
        Returns a new PowerSeriesRing.

        base_ring - a commutative ring
        precision_cap - an integer giving the precision for the prefix-adic coefficients
        ngens - the number of variables
        weights - a tuple of *positive* Python integer weights, one for each variable
        prefix - a string giving the prefix for the variables
        names - a list of variable names; this overrides prefix if given

        """
        self.base_ring = base_ring
        self.precision_cap = precision_cap
        self.ngens = ngens
        self._prefix = prefix
        self._names = [prefix + str(i) for i in range(ngens)]

        if weights is None:
            self.weights = [1 for i in range(ngens)]
        else:
            if len(weights) != ngens:
                raise ValueError(
                    f"The number {len(weights)} is not equal to the number of generators {ngens}."
                )
            for i in weights:
                if i <= 0:
                    raise ValueError("Weights must be positive.")
            self.weights = weights

        # Initialize one and zero since these are used so often.
        AbstractRing.__init__(self, Series, exact=base_ring.exact)

        self._polynomial_ring = PolynomialRing(
            base_ring=self.base_ring,
            ngens=self.ngens,
            prefix=self._prefix,
            packed=False,
        )

        self.one = self(1)
        self.zero = self(0)

        self.gens = []
        for generator in self._polynomial_ring.gens:
            # We have to coerce the SAGE integer into a Python integer.
            self.gens.append(self._unflatten(generator))

    def __call__(self, data):
        if self._polynomial_ring._special:
            self._polynomial_ring = self._polynomial_ring.to_generic()
            # if isinstance(data, int):
            #     if data == 0:
            #         return Series(self, SeriesData(self.base_ring, [], self.precision_cap))
            #     else:
            #         return Series(
            #             self,
            #             SeriesData(
            #                 self.base_ring,
            #                 [
            #                     (
            #                         0,
            #                         self._polynomial_ring.element_class.data_class(
            #                             data,
            #                             self._polynomial_ring.ctx
            #                         ),
            #                     ),
            #                 ],
            #                 self.precision_cap,
            #             ),
            #         )

            # if isinstance(data, self.base_ring.element_class):
            #     return Series(
            #         self,
            #         SeriesData(
            #             self.base_ring,
            #             [
            #                 (
            #                     0,
            #                     self._polynomial_ring.element_class.data_class(
            #                         self.base_ring,
            #                         {
            #                             self._polynomial_ring._monomial_class.from_sparse_tuple(
            #                                 tuple()
            #                             ): data.data
            #                         },
            #                     ),
            #                 ),
            #             ],
            #             self.precision_cap,
            #         ),
            #     )

            # if isinstance(data, self.base_ring.element_class.data_class):
            #     return Series(
            #         self,
            #         SeriesData(
            #             self.base_ring,
            #             [
            #                 (
            #                     0,
            #                     self._polynomial_ring.element_class.data_class(
            #                         self.base_ring,
            #                         {
            #                             self._polynomial_ring._monomial_class.from_sparse_tuple(
            #                                 tuple()
            #                             ): data
            #                         },
            #                     ),
            #                 ),
            #             ],
            #             self.precision_cap,
            #         ),
            #     )

            # return Series(self, data)

        if isinstance(data, int):
            if data == 0:
                return Series(self, SeriesData(self.base_ring, [], self.precision_cap))
            else:
                return Series(
                    self,
                    SeriesData(
                        self.base_ring,
                        [
                            (
                                0,
                                self._polynomial_ring.element_class.data_class(
                                    self.base_ring,
                                    {
                                        self._polynomial_ring._monomial_class.from_sparse_tuple(
                                            tuple()
                                        ): self.base_ring(
                                            data
                                        ).data
                                    },
                                ),
                            ),
                        ],
                        self.precision_cap,
                    ),
                )

        if isinstance(data, self.base_ring.element_class):
            return Series(
                self,
                SeriesData(
                    self.base_ring,
                    [
                        (
                            0,
                            self._polynomial_ring.element_class.data_class(
                                self.base_ring,
                                {
                                    self._polynomial_ring._monomial_class.from_sparse_tuple(
                                        tuple()
                                    ): data.data
                                },
                            ),
                        ),
                    ],
                    self.precision_cap,
                ),
            )

        if isinstance(data, self.base_ring.element_class.data_class):
            return Series(
                self,
                SeriesData(
                    self.base_ring,
                    [
                        (
                            0,
                            self._polynomial_ring.element_class.data_class(
                                self.base_ring,
                                {
                                    self._polynomial_ring._monomial_class.from_sparse_tuple(
                                        tuple()
                                    ): data
                                },
                            ),
                        ),
                    ],
                    self.precision_cap,
                ),
            )

        return Series(self, data)

    def _flatten(self, element):
        """
        Returns the element as an element in the underlying polynomial ring.
        """
        if element.ring != self:
            raise TypeError("Input must be a member of self.")
        out = self._polynomial_ring.zero
        for _, coeff in element.term_list:
            out += coeff
        return out

    def _unflatten_data(self, flat_polynomial_data):
        """
        Takes a PolynomialData instance and returns a SeriesData instance.
        """
        out_degrees = {}
        for m, c in flat_polynomial_data.monomial_dictionary.items():
            deg = 0
            for i in range(len(m.degrees) // 2):
                deg += self.weights[2 * i] * m.degrees[2 * i + 1]
            if deg < self.precision_cap:
                if deg in out_degrees:
                    out_degrees[deg] += self._polynomial_ring.element_class.data_class(
                        self.base_ring, {m: c}
                    )
                else:
                    out_degrees[deg] = self._polynomial_ring.element_class.data_class(
                        self.base_ring, {m: c}
                    )
        out_degrees_list = list(out_degrees.keys())
        out_degrees_list.sort()
        term_list = []
        for deg in out_degrees_list:
            term_list.append((deg, out_degrees[deg]))
        return self.element_class.data_class(
            self.base_ring, term_list, self.precision_cap
        )

    def _unflatten(self, flat_polynomial):
        """
        Takes an element of self._polynomial_ring and unpacks it into a power series,
        ignoring terms of degree at or above the precision cap.
        """
        return self.element_class(self, self._unflatten_data(flat_polynomial.data))

    def __str__(self):
        return (
            f"Absolute precision power series ring in {self._names} with weights {self.weights} "
            f"and with absolutely capped precision {self.precision_cap} over {self.base_ring}"
        )

    def __repr__(self):
        return self.__str__()


class PowerSeriesRingMorphism(AbstractRingMorphism):
    """
    Homomorphisms of weighted power series ring.
    """

    def __init__(self, *, domain, codomain, coefficient_morphism, action_on_generators):
        if (
            domain.base_ring != coefficient_morphism.domain
            or codomain.base_ring != coefficient_morphism.codomain
        ):
            raise TypeError(
                "The coefficient homomorphism is not compatible with the given domain and codomain."
            )
        if len(action_on_generators) != domain.ngens:
            raise TypeError(
                "Cannot interpret {} as a homomorphism from {} to {}".format(
                    action_on_generators, domain, codomain
                )
            )
        if codomain.precision_cap > domain.precision_cap:
            raise ValueError(
                "Codomain precision cap is larger than domain precision cap."
            )
        for i in range(len(action_on_generators)):
            if action_on_generators[i].ring != codomain:
                raise TypeError(
                    "The elements of 'action_on_generators' are not members of the codomain."
                )

        self.coefficient_morphism = coefficient_morphism
        self.domain = domain
        self.codomain = codomain
        self.action_on_generators = action_on_generators

    @functools.cache
    def _call_on_generator_power(self, idx, e):
        return self.action_on_generators[idx] ** ZZ(e)

    @functools.cache
    def _call_on_monomial(self, t):
        new_term = self.codomain.one
        for j in range(len(t.degrees) // 2):
            new_term *= self._call_on_generator_power(
                t.degrees[2 * j], t.degrees[2 * j + 1]
            )
        return new_term

    def __call__(self, f):
        # TODO: this assumes that the input and output of
        # self.coefficient_morphism consist of elements of the data classes of
        # the base_rings.
        if isinstance(f, MonomialData):
            return self._call_on_monomial(f)

        x = self.codomain.zero
        for m, c in f.term_data():
            x += self.coefficient_morphism(self.domain(c)) * self._call_on_monomial(m)
        return x

    def __str__(self):
        return f"Homomorphism from {self.domain} to {self.codomain} defined by {self.action_on_generators} on generators."

    def __repr__(self):
        return self.__str__()



"""
Module constants for JACAMAR.
"""

import flint

CONST_E = flint.arb.const_e()
PI = flint.arb.pi()
CONST_LOG2 = flint.arb.const_log2()
CONST_EULER = flint.arb.const_euler()
PACKING_BOUND = 2**16
MATRIX_SWITCH = 249  # TODO: determine this number during build via tests



"""
TEST_INTEGERS

Tests for the Integer, IntegerPython, IntegerRing, and IntegerRingPython
classes.
"""

import pytest
from flint import fmpz
from flint.utils.flint_exceptions import DomainError
from jacamar.rings.integers import (
    ZZ,
    ZZ_py,
    Integer,
    IntegerPython,
    gcd,
    factor,
    factorial,
)


class TestInteger:
    """Tests for the Integer class."""

    def test_rings(self):
        assert ZZ(6027945939101000).ring == ZZ
        assert ZZ_py(6027945939101000).ring == ZZ_py

    def test_classes(self):
        assert Integer.data_class == fmpz
        assert IntegerPython.data_class == int

    def test_equality_among_the_classes(self):
        """Tests that ZZ(n) == ZZ_py(n) == Integer(n) == Integer_py(n)."""
        with pytest.raises(AttributeError):
            fmpz(159728757) == ZZ(159728757)
        assert ZZ(159728757) == ZZ_py(159728757)
        assert ZZ(159728757) == Integer(ZZ, 159728757)
        assert ZZ_py(159728757) == IntegerPython(ZZ_py, 159728757)
        with pytest.raises(AttributeError):
            159728757 == ZZ(159728757)
        assert ZZ(159728757).data == fmpz(159728757)

    def test_add(self):
        """Tests __add__."""
        assert ZZ(61146125) + ZZ(98582632) == ZZ(159728757)
        assert ZZ_py(61146125) + ZZ_py(98582632) == ZZ_py(159728757)
        with pytest.raises(AttributeError):
            ZZ(7) + 7

    def test_iadd(self):
        """Tests __iadd__."""
        x = ZZ(61146125)
        x += ZZ(98582632)
        assert x == ZZ(159728757)

        x_py = ZZ_py(61146125)
        x_py += ZZ_py(98582632)
        assert x == ZZ_py(159728757)

    def test_sub(self):
        """Tests __sub__."""
        assert ZZ(61146125) - ZZ(98582632) == ZZ(-37436507)
        assert ZZ_py(61146125) - ZZ_py(98582632) == ZZ_py(-37436507)
        with pytest.raises(AttributeError):
            ZZ(7) - 7

    def test_isub(self):
        """Tests __isub__."""
        x = ZZ(61146125)
        x -= ZZ(98582632)
        assert x == ZZ(-37436507)

        x_py = ZZ_py(61146125)
        x_py -= ZZ_py(98582632)
        assert x == ZZ_py(-37436507)

    def test_neg(self):
        """Tests __neg__."""
        assert -ZZ(98582632) == ZZ(-98582632)
        assert -ZZ_py(98582632) == ZZ_py(-98582632)

    def test_mul(self):
        """Tests __mul__."""
        assert ZZ(61146125) * ZZ(98582632) == ZZ(6027945939101000)
        assert ZZ_py(61146125) * ZZ_py(98582632) == ZZ_py(6027945939101000)
        with pytest.raises(AttributeError):
            ZZ(7) * 7

    def test_imul(self):
        """Tests __imul__."""
        x = ZZ(61146125)
        x *= ZZ(98582632)
        assert x == ZZ(6027945939101000)

        x_py = ZZ_py(61146125)
        x_py *= ZZ_py(98582632)
        assert x == ZZ_py(6027945939101000)

    def test_truediv_notimplemented(self):
        """Tests that x/y returns NotImplemented."""
        with pytest.raises(TypeError):
            ZZ(7) / ZZ(7)
        with pytest.raises(TypeError):
            ZZ_py(7) / ZZ_py(7)

    def test_itruediv_notimplemented(self):
        """Tests that x /= y returns NotImplemented."""
        with pytest.raises((DomainError, TypeError)):
            x = ZZ(7)
            x /= ZZ(7)
        with pytest.raises(TypeError):
            x_py = ZZ_py(7)
            x_py /= ZZ_py(7)

    def test_divmod(self):
        """Tests integer division and remainder."""
        assert divmod(ZZ(98582632), ZZ(61146125)) == (ZZ(1), ZZ(37436507))
        assert divmod(ZZ_py(98582632), ZZ_py(61146125)) == (ZZ_py(1), ZZ_py(37436507))
        with pytest.raises(TypeError):
            ZZ(7) / 7

    def test_floordiv(self):
        """Tests the integer __floordiv__ method."""
        assert ZZ(7) // ZZ(2) == ZZ(3)
        assert ZZ_py(7) // ZZ_py(2) == ZZ_py(3)
        assert ZZ(7) // ZZ_py(2) == ZZ_py(3)
        with pytest.raises(AttributeError):
            ZZ(7) // 2

    def test_ifloordiv(self):
        """Tests __ifloordiv__."""
        x = ZZ(7)
        x //= ZZ(2)
        assert x == ZZ(3)

        x_py = ZZ_py(7)
        x_py //= ZZ_py(2)
        assert x == ZZ_py(3)

    def test_mod(self):
        """Tests __mod___."""
        assert ZZ(11) % ZZ(7) == ZZ(4)
        assert ZZ_py(11) % ZZ_py(7) == ZZ_py(4)
        with pytest.raises(AttributeError):
            ZZ(7) % 3

    def test_imod(self):
        """Tests __imod__."""
        x = ZZ(11)
        x %= ZZ(7)
        assert x == ZZ(4)

        x_py = ZZ_py(11)
        x_py %= ZZ_py(7)
        assert x == ZZ_py(4)

    def test_pow(self):
        """Tests powering."""
        with pytest.raises(AttributeError):
            ZZ(415) ** 6
        assert ZZ(415) ** ZZ(6) == ZZ(5108443333890625)
        assert ZZ_py(415) ** ZZ_py(6) == ZZ_py(5108443333890625)

    def test_ipow(self):
        """Tests __ipow__."""
        x = ZZ(415)
        x **= ZZ(6)
        assert x == ZZ(5108443333890625)

        x_py = ZZ_py(415)
        x_py **= ZZ_py(6)
        assert x == ZZ_py(5108443333890625)

    def test_mixing(self):
        """Test correct type of output."""
        assert isinstance(ZZ(7) + ZZ_py(9), IntegerPython)
        assert isinstance(ZZ(7) * ZZ_py(9), IntegerPython)
        assert isinstance(ZZ(7) - ZZ_py(9), IntegerPython)
        assert isinstance(ZZ(7) ** ZZ_py(9), Integer)
        assert isinstance(ZZ(11) % ZZ_py(4), Integer)
        assert isinstance(divmod(ZZ(11), ZZ_py(4))[0], Integer)
        assert isinstance(ZZ(7) // ZZ_py(2), Integer)

        assert isinstance(ZZ_py(7) + ZZ(9), Integer)
        assert isinstance(ZZ_py(7) * ZZ(9), Integer)
        assert isinstance(ZZ_py(7) - ZZ(9), Integer)
        assert isinstance(ZZ_py(7) ** ZZ(9), IntegerPython)
        assert isinstance(ZZ_py(11) % ZZ(4), IntegerPython)
        assert isinstance(divmod(ZZ_py(11), ZZ(4))[0], IntegerPython)
        assert isinstance(ZZ_py(7) // ZZ(2), IntegerPython)

    def test_prime(self):
        """Tests is_prime."""
        assert ZZ(104729).is_prime()
        assert not ZZ(104727).is_prime()

    def test_str(self):
        """Tests __str__ methods."""
        assert ZZ(104729).__str__() == "104729"
        assert ZZ_py(104729).__str__() == "104729"
        assert ZZ.__str__() == "The ring of Integers (via flint.fmpz)."
        assert ZZ_py.__str__() == "The ring of Integers (via Python's int)."

    def test_repr(self):
        """Tests __repr__ methods."""
        assert ZZ(104729).__repr__() == "104729"
        assert ZZ_py(104729).__repr__() == "104729"
        assert ZZ.__repr__() == "The ring of Integers (via flint.fmpz)."
        assert ZZ_py.__repr__() == "The ring of Integers (via Python's int)."

    def test_abs(self):
        """Tests __abs__ method."""
        assert abs(ZZ(-5)) == ZZ(5)
        assert abs(ZZ_py(-5)) == ZZ_py(5)

    def test_eq(self):
        """Tests the __eq__ method."""
        assert ZZ(5) == ZZ_py(5)
        assert not ZZ(5) == ZZ(6)
        assert not ZZ_py(5) == ZZ_py(6)
        with pytest.raises(AttributeError):
            ZZ(5) == 5
        with pytest.raises(AttributeError):
            5 == ZZ(5)

    def test_ge(self):
        """Tests the __ge__ method."""
        assert ZZ(5) >= ZZ(5)
        assert ZZ(5) >= ZZ(-13)
        assert ZZ(5) >= ZZ_py(0)
        assert not ZZ(5) >= ZZ(7)
        with pytest.raises(AttributeError):
            ZZ(5) >= 5

    def test_gt(self):
        """Tests the __gt__ method."""
        assert not ZZ(5) > ZZ(5)
        assert ZZ(5) > ZZ(-13)
        assert ZZ(5) > ZZ_py(0)
        assert not ZZ(5) > ZZ(7)
        with pytest.raises(AttributeError):
            ZZ(5) > 5

    def test_le(self):
        """Tests the __le__ method."""
        assert ZZ(5) <= ZZ(5)
        assert not ZZ(5) <= ZZ(-13)
        assert not ZZ(5) <= ZZ_py(0)
        assert ZZ(5) <= ZZ(7)
        with pytest.raises(AttributeError):
            ZZ(5) <= 5

    def test_lt(self):
        """Tests the __lt__ method."""
        assert not ZZ(5) < ZZ(5)
        assert not ZZ(5) < ZZ(-13)
        assert not ZZ(5) < ZZ_py(0)
        assert ZZ(5) < ZZ(7)
        with pytest.raises(AttributeError):
            ZZ(5) < 5

    def test_ne(self):
        """Tests the __ne__ method."""
        assert ZZ(5) != ZZ(6)
        assert ZZ_py(5) != ZZ_py(6)
        assert not ZZ(5) != ZZ_py(5)
        with pytest.raises(AttributeError):
            ZZ(5) != 5


class test_combinations:
    """Tests the IntegerRing.combinations class method."""
    assert ZZ.combinations(10,3) == ZZ(120)
    assert ZZ.binomial(10,3) == ZZ(120)
    assert ZZ.binomial(10,0) == ZZ(1)
    assert ZZ.binomial(10,11) == ZZ(0)
    with pytest.raises(OverflowError):
        ZZ.binomial(10,-3)


class TestIntegerFunctions:
    """Tests FLINT ZZ functions."""

    def test_gcd(self):
        """Tests the gcd global function."""
        assert gcd(ZZ(15), ZZ(6)) == ZZ(3)

    def test_factor(self):
        """Tests the factor function."""
        assert factor(ZZ(12)) == [(ZZ(2), ZZ(2)), (ZZ(3), ZZ(1))]

    def test_factorial(self):
        """Tests the factorial function."""
        assert factorial(ZZ(5)) == ZZ(120)



"""
POLYNOMIALS

Classes for generic polynomial rings.

Currently, there are two supported exponent types: PackedMonomialData and
SparseMonomialData.

The former are packed into a Python `int`. The largest power supported is given
by the module-level constant PACKING_BOUND.

The latter are designed based on the `ETuple` class from `SAGE`. In particular,
they are sparse. In a polynomial ring on x0, x1, x2, x3, a monomial
x0*x2^3*x3^7 would be encoded as (0,1,2,3,3,7), where the even indices indicate
the present variables and the odd indices indicate their powers. The empty
tuple () represents the monomial of 1.

AUTHORS:
- Benjamin Antieau (2024): initial version.
"""

import functools
import flint
import pyximport

pyximport.install()
import jacamar.rings.cpoly as cpoly
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.rings import AbstractRing
from jacamar.rings.integers import ZZ
from jacamar.rings.rationals import QQ
from jacamar.rings.morphisms import AbstractRingMorphism
from jacamar.constants import PACKING_BOUND

# The following constant controls the maximum allowed weight of a power x^n in a
# monomial: PACKING_BOUND = 2 ** 16.


class MonomialData:
    """Abstract class for monomials."""

    def __init__(self):
        pass

    @classmethod
    def from_packed_integer(cls, _):
        """Construct from an integer."""
        return NotImplemented

    @classmethod
    def from_sparse_tuple(cls, _):
        """Construct from a sparse tuple."""
        return NotImplemented

    @classmethod
    def from_tuple(cls, _):
        """Construct from a tuple."""
        return NotImplemented


class PackedMonomialData(MonomialData):
    """
    A class for packing monomial data into a single integer.
    """

    def __init__(self, n):
        self.weight = n
        MonomialData.__init__(self)

    @classmethod
    def from_packed_integer(cls, n):
        return cls(n)

    @classmethod
    def from_sparse_tuple(cls, t):
        x = 0
        for i in range(len(t) // 2):
            x += t[2 * i + 1] * (PACKING_BOUND ** t[2 * i])
        return cls(x)

    @classmethod
    def from_tuple(cls, t):
        x = 0
        for i, pwr in enumerate(t):
            x += pwr * (PACKING_BOUND**i)
        return cls(x)

    @functools.cached_property
    def degrees(self):
        """Returns the sparse tuple representation of self."""
        count = 0
        x = []
        n = self.weight
        while n > 0:
            q, r = divmod(n, PACKING_BOUND)
            if r > 0:
                x.extend((count, r))
            n = q
            count += 1
        return tuple(x)

    def __hash__(self):
        return self.weight.__hash__()

    def __eq__(self, other):
        return self.weight == other.weight

    def __mul__(self, other):
        """
        Multiplication of monomials.

        WARNING: this could result in nonsense if there is an overflow
        encountered. Ensure that no variable is powered to more than the
        module-level constant PACKING_BOUND.
        """
        return other.__class__(cpoly.c_packed_mul(self.weight, other.weight))

    def __str__(self):
        return str(self.weight)

    def __repr__(self):
        return self.__str__()


class SparseMonomialData(MonomialData):
    """
    The class of a monomial in a PolynomialData.

    INPUT:
    - `t`     -- a tuple of Python integers
    """

    def __init__(self, *t):
        t: tuple
        if len(t) == 1 & isinstance(t, tuple):
            self.degrees = t[0]
        else:
            self.degrees = tuple(t)

    @classmethod
    def from_packed_integer(cls, n):
        count = 0
        x = []
        while n > 0:
            q, r = divmod(n, PACKING_BOUND)
            if r > 0:
                x.extend((count, r))
            n = q
            count += 1
        return cls(tuple(x))

    @classmethod
    def from_sparse_tuple(cls, t):
        return cls(t)

    @classmethod
    def from_tuple(cls, t):
        x = []
        for i, pwr in enumerate(t):
            if pwr != 0:
                x.extend((i, pwr))
        return cls(tuple(x))

    def __hash__(self):
        return hash(self.degrees)

    def __str__(self):
        return str(self.degrees)

    def __repr__(self):
        return str(self)

    def __mul__(self, other):
        return cpoly.c_sparse_mul(self, other)

    def __eq__(self, other):
        return self.degrees == other.degrees


class PolynomialData:
    """
    The underlying data class of polynomials.
    The `data` input should be a dictionary `{m:c}` where `m` is a `MonomialData`
    and `c` is an element of the data_class of the element_class of the `base_ring`.
    Each operation returns a PolynomialData class with no terms with
    coefficient equal to zero if the inputs have this property.

    In practice, the methods in this class work only under the assumption that
    the keys are of a fixed type of MonomialData. It is not designed so that
    the types can be mixed.
    """

    def __init__(self, base_ring, monomial_dictionary):
        self.base_ring = base_ring
        self.monomial_dictionary = {}
        for m, c in monomial_dictionary.items():
            if c != self.base_ring.zero.data:
                self.monomial_dictionary[m] = c

    def is_zero(self):
        """Tests if self is zero."""
        if len(self.monomial_dictionary) == 0:
            return True
        return False

    def term_data(self):
        """Returns the iterator over the items {m:c} in self.monomial_dictionary."""
        return self.monomial_dictionary.items()

    def __str__(self):
        return str(self.monomial_dictionary)

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        if len(self.monomial_dictionary) < len(other.monomial_dictionary):
            new_dict = other.monomial_dictionary.copy()
            other_dict = self.monomial_dictionary

        else:
            new_dict = self.monomial_dictionary.copy()
            other_dict = other.monomial_dictionary

        for m, c in other_dict.items():
            try:
                new_dict[m] += c
            except KeyError:
                new_dict[m] = c
        return other.__class__(other.base_ring, new_dict)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        new_dict = self.monomial_dictionary.copy()
        for m, c in new_dict.items():
            new_dict[m] = -c
        return self.__class__(self.base_ring, new_dict)

    def __mul__(self, other):
        new_dict = {}

        for m, c in self.monomial_dictionary.items():
            for n, d in other.monomial_dictionary.items():
                k = m * n
                e = c * d
                if k in new_dict:
                    new_dict[k] += e
                else:
                    new_dict[k] = e
        return other.__class__(other.base_ring, new_dict)

    def __rmul__(self, other):
        """
        We just try to multiply the coefficients of self with other.
        """
        new_dict = {}
        for m, c in self.monomial_dictionary.items():
            new_dict[m] = c * other
        return self.__class__(self.base_ring, new_dict)

    def __eq__(self, other):
        return (self.base_ring == other.base_ring) & (
            self.monomial_dictionary == other.monomial_dictionary
        )

    def __pow__(self, n):
        if isinstance(n, (flint.fmpz, int)):
            if n < 0:
                raise TypeError("Cannot power a polynomial by a negative integer.")

            apow = self
            while not n & 1:
                # While even
                apow *= apow
                # Bitshift
                n >>= 1

            # Now multiply together the correct factors a^(2^i)
            res = apow
            n >>= 1
            while n:
                apow *= apow
                if n & 1:
                    res = apow * res
                n >>= 1
            return res

        raise TypeError(f"Cannot power polynomial by {n}.")

    def __call__(self, arg):
        """The PolynomialData class can be called on a monomial."""
        try:
            return self.monomial_dictionary[arg]
        except KeyError:
            return self.base_ring.zero.data

    def evaluate(self, args):
        """Returns f(a_1,...,a_n) where arguments=[a_1,...,a_n]."""
        x = self.base_ring.zero.data
        for key, value in self.monomial_dictionary.items():
            monomial_part = self.base_ring.one.data
            for i in range(len(key.degrees) // 2):
                monomial_part *= args[key.degrees[2 * i]] ** key.degrees[2 * i + 1]
            x += value * monomial_part
        return x


class Polynomial(AbstractRingElement):
    """
    The ring of coefficients is `self.base_ring` while the ambient polynomial
    ring is `self.ring`.
    """

    data_class = PolynomialData

    def __init__(self, ring, data):
        self.base_ring = ring.base_ring
        self.data = data

        AbstractRingElement.__init__(
            self,
            ring,
            data,
        )

    def term_data(self):
        """Returns the items of the underlying monomial dictionary."""
        return self.data.term_data()

    def __mul__(self, other):
        if self.ring == other.ring:
            return other.__class__(other.ring, self.data * other.data)
        # Else, coerce self into the other's ring and multiply.
        return other.ring(self) * other

    def __call__(self, *args):
        """
        Evaluate self at x. If x is an instance of MonomialData, returns the
        coefficient. If x is a tuple of elements, then evaluate on these.
        """
        if len(args) == 1:
            if isinstance(args[0], MonomialData):
                return self.base_ring(self.data(args[0]))

        if len(args) == self.ring.ngens:
            try:
                return self.base_ring(self.data.evaluate([x.data for x in args]))
            except AttributeError:
                pass

            # Fall back on making it into a monomial and accessing its
            # coefficient.
            m = self.ring._monomial_class.from_tuple(tuple(args))
            return self.base_ring(self.data(m))

        return NotImplemented

    def __str__(self):
        """String representation."""
        if len(self.data.monomial_dictionary) == 0:
            return "0"
        keys = list(self.data.monomial_dictionary.keys())

        # Parse the first key
        key = keys[0]
        value = self.data.monomial_dictionary[key]
        try:
            bl = value < self.base_ring.zero.data
        except TypeError:
            bl = bool(0)
        if bl:
            token = str(-self.base_ring(value))
        else:
            token = str(self.base_ring(value))
        if token.count(" ") > 0:
            term = "(" + token + ")"
        else:
            if bl:
                term = "- " + token
            else:
                term = token

        for j in range(len(key.degrees) // 2):
            term += "*"
            term += self.ring._prefix
            term += str(key.degrees[2 * j])
            term += f"^{key.degrees[2*j+1]}"
        return_string = term

        # Parse the latter keys
        for i in range(1, len(keys)):
            key = keys[i]
            value = self.data.monomial_dictionary[key]
            # Parse the key
            try:
                bl = value < self.base_ring.zero.data
            except TypeError:
                bl = bool(0)
            if bl:
                token = str(-self.base_ring(value))
            else:
                token = str(self.base_ring(value))
            if token.count(" ") > 0:
                term = " + (" + token + ")"
            else:
                if bl:
                    term = " - " + token
                else:
                    term = " + " + token
            for j in range(len(key.degrees) // 2):
                term += "*"
                term += self.ring._prefix
                term += str(key.degrees[2 * j])
                term += f"^{key.degrees[2*j+1]}"
            return_string += term
        return return_string

    def __repr__(self):
        """String representation."""
        return str(self)


class SpecialPolynomial(AbstractRingElement):
    """
    The ring of coefficients is `self.base_ring` while the ambient polynomial
    ring is `self.ring`.
    """

    # data_class = flint.fmpz_mpoly

    def __init__(self, ring, data):
        self.base_ring = ring.base_ring

        AbstractRingElement.__init__(
            self,
            ring,
            data,
        )

    def term_data(self):
        """Returns the items of the underlying monomial dictionary."""
        return self.data.to_dict().items()

    def __mul__(self, other):
        if self.ring == other.ring:
            return other.__class__(other.ring, self.data * other.data)
        # Else, coerce self into the other's ring and multiply.
        if self.ring.ctx == other.ring.ctx:
            return other.ring(self) * other
        raise ValueError(
            f"Polynomials exist in different contexts. {self.ring.ctx} != {other.ring.ctx}"
        )

    def __call__(self, *args):
        """
        Evaluate self at x. If x is an instance of MonomialData, returns the
        coefficient. If x is a tuple of elements, then evaluate on these.
        """
        if len(args) == self.ring.ngens:
            return self.base_ring(self.data(*tuple(x.data for x in args)))

        return NotImplemented

    def __str__(self):
        """String representation."""
        return str(self.data)

    def __repr__(self):
        """String representation."""
        return str(self)


class PolynomialRing(AbstractRing):
    """
    Class for generic (multivariable) polynomial rings.

    INPUT:
    - `base_ring`   -- the FLINT ring from which the coefficients are taken
    - `ngens`       -- the numberg of generators
    - `prefix`      -- a string such as 'x' from which the generators x0,x1,... are named
    - `weights`     -- a list of `ngens` weights; if `None`, these default to 1; they are Python integers
    """

    element_class = Polynomial

    def __init__(
        self, *, base_ring, ngens, prefix, weights=None, packed=True, special=True
    ):
        # Choose constructuion method: fmpz_mpoly works only for ZZ as of now
        if base_ring == ZZ and special:
            self.element_class = SpecialPolynomial
            self._special = True
            self._data_class = flint.fmpz_mpoly
            self._context_class = flint.fmpz_mpoly_ctx
            self.element_class.data_class = flint.fmpz_mpoly

        elif base_ring == QQ and special:
            self.element_class = SpecialPolynomial
            self._special = True
            self._data_class = flint.fmpq_mpoly
            self._context_class = flint.fmpq_mpoly_ctx
            self.element_class.data_class = flint.fmpq_mpoly

        else:
            self._special = False

        # Special construction
        if self._special:
            self.base_ring = base_ring
            self.ngens = ngens
            if packed:
                self._monomial_class = PackedMonomialData
            else:
                self._monomial_class = SparseMonomialData
            self.packed = packed
            self._prefix = prefix
            self._names = [prefix + str(i) for i in range(ngens)]
            if weights is None:
                self.weights = [1 for i in range(ngens)]
            else:
                if len(weights) != ngens:
                    raise ValueError(
                        f"The number {len(weights)} is not equal to the number of generators {ngens}."
                    )
                self.weights = weights

            self.gens = []
            ctx = self._context_class.get(
                [f"{self._prefix}{x}" for x in range(self.ngens)], flint.Ordering.lex
            )
            self.ctx = ctx
            for i in range(self.ngens):
                self.gens.append(
                    SpecialPolynomial(
                        self,
                        self._data_class(
                            {
                                tuple(
                                    int(x == i) for x in range(self.ngens)
                                ): self.base_ring.one.data
                            },
                            ctx,
                        ),
                    )
                )

            AbstractRing.__init__(self, SpecialPolynomial, exact=base_ring.exact)

            # Initialize one and zero since these are used so often.
            self.one = self(1)
            self.zero = self(0)

        # Generic construction
        else:
            self.base_ring = base_ring
            self.ngens = ngens
            if packed:
                self._monomial_class = PackedMonomialData
            else:
                self._monomial_class = SparseMonomialData
            self.packed = packed
            self._prefix = prefix
            self._names = [prefix + str(i) for i in range(ngens)]
            if weights is None:
                self.weights = [1 for i in range(ngens)]
            else:
                if len(weights) != ngens:
                    raise ValueError(
                        f"The number {len(weights)} is not equal to the number of generators {ngens}."
                    )
                self.weights = weights

            self.gens = []
            for i in range(self.ngens):
                self.gens.append(
                    Polynomial(
                        self,
                        PolynomialData(
                            self.base_ring,
                            {
                                self._monomial_class.from_sparse_tuple(
                                    (i, 1)
                                ): self.base_ring.one.data
                            },
                        ),
                    )
                )

            AbstractRing.__init__(self, Polynomial, exact=base_ring.exact)

            # Initialize one and zero since these are used so often.
            self.one = self(1)
            self.zero = self(0)

    def __str__(self):
        """String representation."""
        return f"Ring of polynomials in {self.ngens} variables {self._names} with weights {self.weights} over {self.base_ring}"

    def __repr__(self):
        """String representation."""
        return self.__str__()

    def __call__(self, data):
        """Create a new polynomial from data, if possible."""
        # Special calls
        if self._special:
            if isinstance(data, int):
                return SpecialPolynomial(
                    self,
                    self._data_class(data, self.ctx),
                )
            if isinstance(data, dict):
                new_dict = {}
                for key, value in data.items():
                    new_dict[key] = value.data
                return SpecialPolynomial(
                    self,
                    self._data_class(new_dict, self.ctx),
                )
            if isinstance(data, self.element_class):
                if data.ring == self:
                    # Create a new element with the same data.
                    return self.element_class(
                        self, self._data_class(data.data, self.ctx)
                    )
                return self.element_class(
                    self, self._data_class(data.data, data.ring.ctx)
                )
            if isinstance(data, self.element_class.data_class):
                return self.element_class(self, data)
            if isinstance(data, self.base_ring.element_class):
                if data.ring == self.base_ring:
                    return SpecialPolynomial(
                        self,
                        self._data_class(data.data, self.ctx),
                    )

            raise TypeError(
                f"No known constructor for input data of type {type(data)}."
            )

        if isinstance(data, int):
            return Polynomial(
                self,
                PolynomialData(
                    self.base_ring,
                    {
                        self._monomial_class.from_sparse_tuple(()): self.base_ring(
                            data
                        ).data
                    },
                ),
            )
        if isinstance(data, dict):
            new_dict = {}
            for key, value in data.items():
                new_dict[self._monomial_class.from_sparse_tuple(key)] = value.data
            return Polynomial(self, PolynomialData(self.base_ring, new_dict))
        if isinstance(data, self.element_class):
            if data.ring == self:
                # Create a new element with the same data.
                return self.element_class(self, data.data)
        if isinstance(data, self.element_class.data_class):
            if data.base_ring == self.base_ring:
                return self.element_class(self, data)
        if isinstance(data, self.base_ring.element_class):
            if data.ring == self.base_ring:
                return Polynomial(
                    self,
                    PolynomialData(
                        self.base_ring,
                        {
                            self._monomial_class.from_sparse_tuple(()): self.base_ring(
                                data
                            ).data
                        },
                    ),
                )
        raise TypeError(f"No known constructor for input data of type {type(data)}.")

    def to_generic(self):
        """Creates a generic copy of a special polynomial"""
        if not self._special:
            return self
        return self.__class__(
            base_ring=self.base_ring,
            ngens=self.ngens,
            prefix=self._prefix,
            weights=self.weights,
            packed=self.packed,
            special=False,
        )


class PolynomialRingMorphism(AbstractRingMorphism):
    """
    Homomorphisms of polynomial rings.
    """

    def __init__(self, *, domain, codomain, coefficient_morphism, action_on_generators):
        if (
            domain.base_ring != coefficient_morphism.domain
            or codomain.base_ring != coefficient_morphism.codomain
        ):
            raise TypeError(
                "The coefficient homomorphism is not compatible with the given domain and codomain."
            )
        if len(action_on_generators) != domain.ngens:
            raise TypeError(
                f"Cannot interpret {action_on_generators} as a homomorphism from {domain} to {codomain}"
            )
        for gen in action_on_generators:
            if gen.ring != codomain:
                raise TypeError(
                    "The elements of 'action_on_generators' are not members of the codomain."
                )

        self.coefficient_morphism = coefficient_morphism
        self.action_on_generators = action_on_generators
        AbstractRingMorphism.__init__(self, domain, codomain)

    @functools.lru_cache
    def _call_on_generator_power(self, idx, e):
        """Cache the calls to powers of generators."""
        return self.action_on_generators[idx] ** ZZ(e)

    @functools.lru_cache
    def _call_on_monomial(self, t):
        """Cache calls on monomials."""
        new_term = self.codomain.one
        for j in range(len(t.degrees) // 2):
            new_term *= self._call_on_generator_power(
                t.degrees[2 * j], t.degrees[2 * j + 1]
            )
        return new_term

    def __call__(self, f):
        # TODO: this assumes that the input and output of
        # self.coefficient_morphism consist of elements of the data classes of
        # the base_rings.
        if isinstance(f, MonomialData):
            return self._call_on_monomial(f)

        x = self.codomain.zero
        for m, c in f.term_data():
            x += self.coefficient_morphism(
                self.domain.base_ring(c)
            ) * self._call_on_monomial(m)
        return x

    def __str__(self):
        """String representation."""
        return f"Homomorphism from {self.domain} to {self.codomain} defined by {self.action_on_generators} on generators."

    def __repr__(self):
        """String representation."""
        return self.__str__()



"""
TEST_REALS

Tests for the RealNumber and RealRing classes.
"""

import pytest
import flint
from jacamar.rings.reals import (
    RR,
    RR_py,
    sin,
    cos,
    tan,
    asin,
    acos,
    atan,
    csc,
    sec,
    cot,
    abs_lower,
    abs_upper,
    exp,
    factorial,
    log,
    root,
    sqrt,
)
from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.polynomials import (
    SparseMonomialData,
    PolynomialData,
    PolynomialRing,
)
from jacamar.matrices.matrices import Matrix


class TestRealNumber:
    """Tests for the RealNumber class."""

    def test_no_args(self):
        """Tests RR()."""
        assert RR() == RR(0)

    def test_two_args(self):
        """Tests RR(c, r)"""
        assert RR("10.5+/-0.5") == RR("10.4 +/-0.5")

    def test_truediv(self):
        """Tests that x / y."""
        x = RR(1)
        y = RR(3)
        q = x / y
        assert str(q) == "[0.333333333333333 +/- 3.71e-16]"

    def test_itruediv_notimplemented(self):
        """Tests x /= y."""
        x = RR(1)
        y = RR(3)
        x /= y
        assert str(x) == "[0.333333333333333 +/- 3.71e-16]"

    def test_str(self):
        """Tests __str__."""
        assert str(RR(10.25)) == "10.2500000000000"
        assert str(RR) == "The ring of real numbers (via flint.arb)."

    def test_repr(self):
        """Tests __repr__."""
        assert RR.__repr__() == "The ring of real numbers (via flint.arb)."

    def test_mult(self):
        """Test x * y."""
        assert RR(4.0) * RR(2.25) == RR(9.0)

    def test_add(self):
        """Test x + y."""
        assert RR(4.2) + RR(3.5) == RR(7.7)

    def test_sub(self):
        """Test x - y."""
        assert RR(2.1) - RR(3.1) == RR(-1.0)

    def test_add_with_integer(self):
        """Tests __add__ with an integer."""
        assert ZZ(1) + RR(6.765) == RR(7.765)
        with pytest.raises(TypeError):
            RR(6.765) + ZZ(1)

    def test_eq(self):
        """Tests __eq__ with RR and ZZ."""
        assert RR("0 +/- 0.5") == RR(0.4)
        assert RR(0) == ZZ(0)

    def test_mult_with_integer(self):
        """Tests ___mul__ with an integer."""
        assert ZZ(1) * RR(6.765) == RR(6.765)
        with pytest.raises(TypeError):
            RR(6.765) * ZZ(1)

    def test_polynomial(self):
        """Tests creation of real polynomials."""
        r = PolynomialRing(base_ring=RR, ngens=2, prefix="x")
        x0, x1 = r.gens
        f = x0 + x1
        assert f ** ZZ(2) == x0 ** ZZ(2) + ZZ(2) * x0 * x1 + x1 ** ZZ(2)

        r = PolynomialRing(base_ring=RR_py, ngens=2, prefix="x")
        x0, x1 = r.gens
        f = x0 + x1
        assert f ** ZZ_py(2) == x0 ** ZZ_py(2) + ZZ_py(2) * x0 * x1 + x1 ** ZZ_py(2)


class TestRealNumberFunctions:
    """Tests flint RR functions."""

    def test_sin(self):
        """Tests the sin global function."""
        assert str(sin(RR(1))) == "[0.841470984807897 +/- 6.08e-16]"

    def test_cos(self):
        """Tests the cos global function."""
        assert str(cos(RR(1))) == str(flint.arb(1).cos())

    def test_tan(self):
        """Tests the tan function."""
        assert str(tan(RR(1))) == str(flint.arb(1).tan())

    def test_asin(self):
        """Tests the asin global function."""
        assert str(asin(RR(1))) == str(flint.arb(1).asin())

    def test_acos(self):
        """Tests the acos function."""
        assert str(acos(RR(1))) == str(flint.arb(1).acos())

    def test_atan(self):
        """Tests the atan function."""
        assert str(atan(RR(1))) == str(flint.arb(1).atan())

    def test_csc(self):
        """Tests the csc global function."""
        assert str(csc(RR(1))) == str(flint.arb(1).csc())

    def test_sec(self):
        """Tests the sec function."""
        assert str(sec(RR(1))) == str(flint.arb(1).sec())

    def test_cot(self):
        """Tests the cot function."""
        assert str(cot(RR(1))) == str(flint.arb(1).cot())

    def test_abs_upper(self):
        """Tests the abs_upper global function."""
        assert str(abs_upper(RR("1.5+/-1"))) == str(flint.arb("1.5+/-1").abs_upper())

    def test_abs_lower(self):
        """Tests the abs_lower function."""
        assert str(abs_lower(RR("1.5+/-1"))) == str(flint.arb("1.5+/-1").abs_lower())

    def test_factorial(self):
        """Tests the factorial function."""
        assert factorial(RR(5)) == RR(120)

    def test_exp(self):
        """Tests the exp global function."""
        assert str(exp(RR(2))) == str(flint.arb(2).exp())

    def test_log(self):
        """Tests the log function."""
        assert str(log(RR(3.3))) == str(flint.arb(3.3).log())

    def test_root(self):
        """Tests the root function."""
        assert str(root(RR(7), 3)) == str(flint.arb(7).root(3))

    def test_sqrt(self):
        """Tests the sqrt function."""
        assert str(sqrt(RR(4))) == str(flint.arb(4).sqrt())



"""
ELEMENTS

Base classes for elements in commutative rings. The abstract class stores a
concrete `data` attribute, on which it is assumed that the standard arithmetic
operations are implemented.

AUTHORS:
- Benjamin Antieau (2024).
"""


class AbstractRingElement:
    def __init__(self, ring, data):
        self.ring = ring
        # TODO: does this have unexpected copying behavior? It seems like if we
        # change .data in one place it will be changed elsewhere.
        if isinstance(data, self.ring.element_class):
            self.data = data.data
        elif isinstance(data, self.ring.element_class.data_class):
            self.data = data
        else:
            self.data = self.ring.element_class.data_class(data)

    # Arithmetic functions.
    def __add__(self, other):
        """Returns self + other with type that of other."""
        return other.__class__(other.ring, self.data + other.data)

    def __mul__(self, other):
        """Returns self * other with type that of other."""
        return other.__class__(other.ring, self.data * other.data)

    def __sub__(self, other):
        """Returns self - other with type that of other."""
        return other.__class__(other.ring, self.data - other.data)

    def __neg__(self):
        """Returns - self."""
        return self.__class__(self.ring, -self.data)

    def __truediv__(self, other):
        """Returns self / other with type that of self."""
        return NotImplemented

    def __floordiv__(self, other):
        """Returns self // other with type that of self."""
        return self.__class__(self.ring, self.data // other.data)

    def __divmod__(self, other):
        """Returns (self // other, self % other) with types that of self."""
        d = divmod(self.data, other.data)
        return (self.__class__(self.ring, d[0]), self.__class__(self.ring, d[1]))

    def __mod__(self, other):
        """Returns self % other with type that of self."""
        return self.__class__(self.ring, self.data % other.data)

    def __pow__(self, n):
        """Returns self ** n with type that of self."""
        if n == n.ring.zero:
            return self.ring.one
        return self.__class__(self.ring, self.data**n.data)

    # Comparison functions.
    def __abs__(self):
        return self.__class__(self.ring, self.data.__abs__())

    def __eq__(self, other):
        return self.data == other.data

    def __ge__(self, other):
        return self.data >= other.data

    def __gt__(self, other):
        return self.data > other.data

    def __le__(self, other):
        return self.data <= other.data

    def __lt__(self, other):
        return self.data < other.data

    def __ne__(self, other):
        return self.data != other.data



"""
RINGS

A module for commutative rings.

AUTHORS:
- Benjamin Antieau (2024).
"""

from jacamar.rings.morphisms import IdentityRingMorphism


class AbstractRing:
    def __init__(self, element_class, exact=True):
        self.element_class = element_class

        # Whether or not the ring is exact.
        self.exact = exact

        # The following should be overridden for each concrete class.
        self.zero = None
        self.one = None

    def identity_morphism(self):
        return IdentityRingMorphism(self)

    def __call__(self, x):
        return self.element_class(self, x)



"""
C-POLY
Cython implementation of polynomial multiplication methods. 
Compiled in C in the file cpoly.c
"""
import cython
def c_packed_mul(s: int, o: int):
    return s + o

def c_sparse_mul(self, other):
    new_list= []
    self_index: cython.int = 0
    other_index: cython.int = 0
    self_len: cython.int = len(self.degrees)
    other_len: cython.int = len(other.degrees)
    while self_index < self_len and other_index < other_len:
        if self.degrees[self_index] < other.degrees[other_index]:
            new_list.extend(self.degrees[self_index : self_index + 2])
            self_index += 2
        elif self.degrees[self_index] == other.degrees[other_index]:
            x = self.degrees[self_index + 1] + other.degrees[other_index + 1]
            if x != 0:
                new_list.append(self.degrees[self_index])
                new_list.append(x)
            self_index += 2
            other_index += 2
        else:
            new_list.extend(other.degrees[other_index : other_index + 2])
            other_index += 2

    # Now that we've reached the end of at least one list, we add
    # everything from the other list.
    while self_index < self_len:
        new_list.extend(self.degrees[self_index : self_index + 2])
        self_index += 2
    while other_index < other_len:
        new_list.extend(other.degrees[other_index : other_index + 2])
        other_index += 2
    
    return other.__class__(tuple(new_list))

def c_poly_mul(self, other):
    new_dict = {}
    for m, c in self.monomial_dictionary.items():
        for n, d in other.monomial_dictionary.items():
            k = m * n
            e = c * d
            if k in new_dict:
                new_dict[k] += e
            else:
                new_dict[k] = e
    return other.__class__(other.base_ring, new_dict)






"""
INTEGERS

Base classes for ZZ and ZZ_py.

AUTHORS:
- Benjamin Antieau (2024).
"""

import flint
from jacamar.rings.elements import AbstractRingElement
from jacamar.rings.rings import AbstractRing


class Integer(AbstractRingElement):
    data_class = flint.fmpz

    def __init__(self, ring, n):
        AbstractRingElement.__init__(
            self,
            ring,
            n,
        )

    def is_prime(self):
        return bool(self.data.is_prime())

    def is_unit(self):
        return self.data == flint.fmpz(1) or self.data == flint.fmpz(-1)

    def inverse(self):
        if self.data == flint.fmpz(1):
            return self
        elif self.data == flint.fmpz(-1):
            return self
        else:
            raise ValueError("Is not invertible.")

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()


class IntegerPython(AbstractRingElement):
    data_class = int

    def __init__(self, ring, n):
        AbstractRingElement.__init__(
            self,
            ring,
            n,
        )

    def is_unit(self):
        return self.data == -1 or self.data == 1

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()


class IntegerRing(AbstractRing):
    def __init__(self):
        AbstractRing.__init__(self, Integer, exact=True)
        self.one = self(1)
        self.zero = self(0)

    def __str__(self):
        return "The ring of Integers (via flint.fmpz)."

    def __repr__(self):
        return self.__str__()

    def combinations(self, n, k):
        return self.element_class(ring=self,n=flint.fmpz.bin_uiui(n,k))

    binomial = combinations


class IntegerRingPython(AbstractRing):
    def __init__(self):
        AbstractRing.__init__(self, IntegerPython, exact=True)
        self.one = self(1)
        self.zero = self(0)

    def __str__(self):
        return "The ring of Integers (via Python's int)."

    def __repr__(self):
        return self.__str__()


# Create global version of the integer rings.
ZZ = IntegerRing()
ZZ_py = IntegerRingPython()


"""
Select functions of ZZ's based on flint.fmpz attributes. (https://fredrikj.net/python-flint/fmpz.html)
"""


def factorial(x):
    """
    Integer factorial of x.
    """
    return x.ring(x.data_class.fac_ui(x.data))


def gcd(x, y):
    """
    Greatest common factor of x and y.
    """
    return x.ring(x.data.gcd(y.data))


def factor(x):
    """
    Factors of x in polynomial form.
    """
    ret = []
    factors = x.data.factor()
    for f in factors:
        ret.append((x.ring(f[0]), x.ring(f[1])))
    return ret



"""
REALS

Base class for RR.
"""

import flint
from jacamar.rings.elements import AbstractRingElement  # type: ignore
from jacamar.rings.rings import AbstractRing  # type: ignore


class RealNumber(AbstractRingElement):
    """
    Base class for real numbers, built on AbstractRingElement.
    """

    data_class = flint.arb  # type: ignore

    def is_unit(self):
        return self != self.ring.zero

    def inverse(self):
        if self.is_unit():
            return self.ring.one / self
        else:
            raise ValueError("Is not invertible.")

    def __init__(self, ring, x=0):
        AbstractRingElement.__init__(self, ring, x)

    def __truediv__(self, other):
        """Returns self / other with type that of self."""
        return self.__class__(self.ring, self.data / other.data)

    def __itruediv__(self, other):
        """Modifies self in place by self / other."""
        self.data /= other.data
        return self

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, RealNumber):
            if self.ring(self.data.abs_lower()) >= other.ring(other.data.abs_lower()):
                return self.ring(self.data.abs_lower()) <= other.ring(
                    other.data.abs_upper()
                )
            else:
                return self.ring(self.data.abs_upper()) >= other.ring(
                    other.data.abs_lower()
                )
        return self.data == other.data


class RealNumberPython(AbstractRingElement):
    """
    Base class for real numbers, built on AbstractRingElement. Interfaces with python.
    """

    data_class = float

    def is_unit(self):
        return self != self.ring.zero

    def inverse(self):
        if self.is_unit():
            return self.ring.one / self
        else:
            raise ValueError("Is not invertible.")

    def __init__(self, ring, x=0):
        AbstractRingElement.__init__(self, ring, self.data_class(x))

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.__str__()


class RealRing(AbstractRing):
    """
    Base class for real rings, built on AbstractRing.
    """

    def __init__(self):
        AbstractRing.__init__(self, RealNumber, exact=False)
        self.one = self(1)
        self.zero = self(0)

    def __call__(self, *args):
        return RealNumber(self, *args)

    def __str__(self):
        return "The ring of real numbers (via flint.arb)."

    def __repr__(self):
        return self.__str__()


class RealRingPython(AbstractRing):
    """
    Base class for real rings, built on AbstractRing. Interfaces with python.
    """

    def __init__(self):
        AbstractRing.__init__(self, RealNumberPython, exact=False)
        self.one = self(1)
        self.zero = self(0)

    def __call__(self, *args):
        return RealNumberPython(self, *args)

    def __str__(self):
        return "The ring of real numbers (via flint.arb)."

    def __repr__(self):
        return self.__str__()


# Create a global version of the real ring.
RR = RealRing()
RR_py = RealRingPython()


"""
Select functions of RR's based on flint.arb attributes. (https://fredrikj.net/python-flint/arb.html)
"""


def sin(x):
    """
    Sine of x.
    """
    return x.ring(x.data.sin())


def cos(x):
    """
    Cosine of x.
    """
    return x.ring(x.data.cos())


def tan(x):
    """
    Tangent of x.
    """
    return x.ring(x.data.tan())


def asin(x):
    """
    Inverse sine of x.
    """
    return x.ring(x.data.asin())


def acos(x):
    """
    Inverse cosine of x.
    """
    return x.ring(x.data.acos())


def atan(x):
    """
    Inverse tanget of x.
    """
    return x.ring(x.data.atan())


def csc(x):
    """
    Cosecant of x.
    """
    return x.ring(x.data.csc())


def sec(x):
    """
    Secant of x.
    """
    return x.ring(x.data.sec())


def cot(x):
    """
    Cotanget of x.
    """
    return x.ring(x.data.cot())


def abs_lower(x):
    """
    The absolute value of the lower bound of x.
    """
    return x.ring(x.data.abs_lower())


def abs_upper(x):
    """
    The absolute value of the upper bound of x.
    """
    return x.ring(x.data.abs_upper())


def exp(x):
    """
    The constat e raised to the power of x.
    """
    return x.ring(x.data.exp())


def factorial(x):
    """
    Factorial, using the gamma function of x.
    """
    return x.ring(x.data.fac())


def log(x):
    """
    Natural logarithm of x.
    """
    return x.ring(x.data.log())


def root(x, n):
    """nth root of x."""
    return x.ring(x.data.root(n))


def sqrt(x):
    """
    Square root of x.
    """
    return x.ring(x.data.sqrt())



"""
TEST_INTMOD

Tests for the IntegerModN and IntegerModNRing classes.
"""

import pytest
from flint import fmpz, nmod  # pylint: disable=no-name-in-module
from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.intmod import IntegerModN, IntegerModNRing
from jacamar.rings.polynomials import PolynomialRing

r1 = IntegerModNRing(ZZ(24))
r2 = IntegerModNRing(ZZ(173))
r3 = IntegerModNRing(ZZ(214102727))


class TestIntegerModN:
    """Tests for the IntegerModN class."""

    def test_simple(self):
        """Test that modular arithmetic works as expected with our interface"""
        n = ZZ(103)
        r = IntegerModNRing(n)
        x = r(5)
        assert x ** ZZ(3) == r(22)

    def test_simple_2(self):
        """Test that creating an IntegerModN instance directly works"""
        r = IntegerModNRing(ZZ(7))
        x = IntegerModN(r, 5)
        assert r(5) == x

    def test_simple_fail(self):
        """Test that inequivalent expressions are evaluated as inequivalent"""
        n = ZZ(60)
        r = IntegerModNRing(n)
        x = r(5)
        assert x ** ZZ(3) != r(4)

    def test_rings(self):
        """The ring attribute of an element generated by the ring is the same ring"""
        assert r3(129127208515966992384).ring == r3
        assert r2(10).ring != r3

    def test_class_equality(self):
        """Tests that rings/elements are equal for equivalent inputs from different
        data classes"""
        print(isinstance(IntegerModNRing(ZZ(134)).modulus.data, fmpz))
        print(isinstance(IntegerModNRing(134).modulus.data, fmpz))
        # assert IntegerModNRing(ZZ(134)) == IntegerModNRing(134)
        # assert IntegerModNRing(134) == IntegerModNRing(ZZ_py(134))
        assert r2(200) == r2(ZZ(200))
        assert r1(123) == r1(ZZ_py(123))
        assert (IntegerModNRing(ZZ(134)))(3) == (IntegerModNRing(ZZ(134)))(3)

    def test_data_class(self):
        """The data class of IntegerModN is flint.nmod as expected"""
        assert IntegerModN(r3, 3).data_class == nmod

    def test_str(self):
        """Test __str__ method output"""
        assert str(r1) == "The ring of Integers modulo 24."
        assert str(r1(10)) == "10"
        assert str(r1(27)) == "3"

    def test_repr(self):
        """Test __repr__ method output"""
        assert repr(r1) == "The ring of Integers modulo 24."
        assert repr(r1(10)) == "10"
        assert repr(r1(27)) == "3"

    def test_add(self):
        """Tests that the addition operator works correctly."""
        assert r2(30) + r2(35) == r2(65)
        assert r2(1000) + r2(146) == r2(108)

    def test_iadd(self):
        """Tests that the addition assignment operator works correctly."""
        x = r2(69)
        x += r2(21)
        assert x == r2(90)
        x += r2(83)
        assert x == r2(0)

    def test_mul(self):
        """Tests that the multiplication operator works correctly."""
        assert r1(12) * r1(2) == r1(0)
        assert r1(12) * r1(3) == r1(12)

    def test_imul(self):
        """Tests that the multiplication assignment operator works correctly."""
        x = r1(13)
        x *= r1(3)
        assert x == r1(15)

    def test_sub(self):
        """Tests that the subtraction operator works correctly."""
        assert r2(100) - r2(40) == r2(60)
        assert r1(100) - r1(40) == r1(60)
        assert r3(3) - r3(-6) == r3(9)
        assert r1(12) - r1(14) == r1(22)

    def test_isub(self):
        """Tests that the subtraction assignment operator works correctly."""
        x = r2(12)
        x -= r2(12)
        assert x == r2(0)

    def test_neg(self):
        """Tests that negation works correctly."""
        x = r1(5)
        x = -x
        assert x == r1(19)

    def test_truediv(self):
        """Tests that the (true) division operator works correctly."""
        r = IntegerModNRing(ZZ(7))
        assert r(3) / r(3) == r(1)
        assert r(3) / r(2) == r(5)

    def test_itruediv(self):
        """Tests that the (true) division assignment operator works correctly."""
        r = IntegerModNRing(ZZ(7))
        x = r(3)
        x /= r(2)
        assert x == r(5)

    def test_floordiv(self):
        """The floor division operator should not be implemented."""
        with pytest.raises(TypeError):
            x = r1(7) // r1(7)
            return x

    def test_ifloordiv(self):
        """The floor division assignment operator should not be implemented."""
        with pytest.raises(TypeError):
            x = r1(7)
            x //= r1(7)

    def test_divmod(self):
        """The divmod operator should not be implemented."""
        with pytest.raises(TypeError):
            x = divmod(r1(7), r1(7))
            return x

    def test_mod(self):
        """The modulo operator should not be implemented."""
        with pytest.raises(TypeError):
            x = r1(7) % r1(2)
            return x

    def test_imod(self):
        """The modulo assignment operator should not be implemented."""
        with pytest.raises(TypeError):
            x = r1(7)
            x %= r1(7)

    def test_pow(self):
        """Tests that the exponentiation (pow) operator works correctly."""
        assert r1(5) ** ZZ(2) == r1(1)
        assert r1(0) ** ZZ(5) == r1(0)
        assert r1(8) ** ZZ(0) == r1(1)
        assert r1(0) ** ZZ(0) == r1(1)

    def test_ipow(self):
        """Tests that the exponentiation (pow) assignment operator works correctly."""
        x = r3(9)
        x **= ZZ(3)
        assert x == r3(729)

    def test_abs(self):
        """The abs operator should not be implemented."""
        with pytest.raises(AttributeError):
            x = abs(r1(-7))
            return x

    def test_eq(self):
        """Tests that the equality operator works correctly."""
        assert r1(2) == r1(2)
        assert r1(2) == r1(26)
        assert not r1(2) == r1(3)
        with pytest.raises(AttributeError):
            return r1(2) == 2

    def test_ge(self):
        """The greater or equal operator should not be implemented."""
        with pytest.raises(TypeError):
            return r1(7) >= r1(7)

    def test_gt(self):
        """The greater operator should not be implemented."""
        with pytest.raises(TypeError):
            return r1(7) > r1(7)

    def test_le(self):
        """The less or equal operator should not be implemented."""
        with pytest.raises(TypeError):
            return r1(7) <= r1(7)

    def test_lt(self):
        """The less operator should not be implemented."""
        with pytest.raises(TypeError):
            return r1(7) < r1(7)

    def test_ne(self):
        """Tests that the inequality operator works correctly."""
        assert r1(2) != r1(3)
        assert not r1(2) != r1(26)
        assert not r1(2) != r1(2)
        with pytest.raises(AttributeError):
            return r1(2) != 3

    def test_polynomial(self):
        """Tests creation of polynomials with base ring of integers modulo n."""
        r = PolynomialRing(base_ring=r2, ngens=2, prefix="x")
        x0, x1 = r.gens
        f = r2(1) * x0
        assert f ** ZZ(2) == (x0 ** ZZ(2))
        assert f * x1 == x0 * x1
        assert r2(-2) * f == r2(171) * x0



"""
MATRICES

A module for dense matrices. We create a single abstract matrix class which
holds a `data` attribute. We assume that this points to a class on which all
matrix operations can be performed and we overload matrices in that way. We
provide a generic implementation of these operations as `_MatrixGeneric`. Special
examples are provided by `FLINT`.

AUTHORS:
- Benjamin Antieau (2024): initial version.
- Luca Zerega (2024): major revisions.
"""

import flint
import numpy as np

from jacamar.rings.integers import ZZ, ZZ_py
from jacamar.rings.reals import RR, RR_py
from jacamar.rings.complexes import CC
from jacamar.rings.rationals import QQ
from jacamar.constants import MATRIX_SWITCH


class _MatrixGenericData:
    """
    This class must be called with all arguments. It is assumed, but not
    checked, that the shape of `entries` conforms with this. In the case that
    `nrows==0` or `ncols==0`, one should have `entries=[]`.
    """

    def __init__(self, *, base_ring, nrows, ncols, entries):
        self.base_ring = base_ring
        self.nrows = nrows
        self.ncols = ncols
        self.entries = entries

    def det(self):
        """Alias for `determinant` method."""
        return self.determinant()

    def determinant(self):
        if self.nrows != self.nrows:
            raise ValueError("Matrix must be square.")
        if self.nrows == 0:
            return self.base_ring(0)
        if self.nrows == 1:
            return self.entries[0][0]
        entries = self.entries

        def minor(array, i, j):
            c = array
            c = c[:i] + c[i + 1 :]
            for k in range(0, len(c)):
                c[k] = c[k][:j] + c[k][j + 1 :]
            return c

        def det(array, n):
            if n == 1:
                return array[0][0]
            if n == 2:
                return array[0][0] * array[1][1] - array[0][1] * array[1][0]
            det = self.base_ring(base_ring=RR, ngens=3, prefix="x", packed=False)
            det = det.zero
            for i in range(0, n):
                m = minor(array, 0, i)
                det = det + RR((-1) ** i) * array[0][i] * det(m, n - 1)
            return det

        return det(entries, self.nrows)

    def transpose(self):
        """Returns copy of the transposed matrix data."""
        entries = self.entries
        new_entries = []
        for i in range(self.ncols):
            new_entries.append([])
            for j in range(self.nrows):
                new_entries[-1].append(entries[j][i])

        return _MatrixGenericData(
            base_ring=self.base_ring,
            nrows=self.ncols,
            ncols=self.nrows,
            entries=new_entries,
        )

    def size(self):
        """Returns size of a matrix as a tuple (rows, cols)."""
        return (self.nrows, self.ncols)

    def __add__(self, other):
        if self.nrows != other.nrows or self.ncols != other.ncols:
            raise ValueError(
                f"Cannot add matrix of size {self.nrows}x{self.ncols} to matrix of size {other.nrows}x{other.ncols}."
            )

        if self.nrows == 0 or other.nrows == 0:
            return other.__class__(
                base_ring=other.base_ring,
                nrows=other.nrows,
                ncols=other.ncols,
                entries=[],
            )
        new_entries = [self.entries[i].copy() for i in range(self.nrows)]
        for i in range(self.nrows):
            for j in range(self.ncols):
                new_entries[i][j] += other.entries[i][j]

        return other.__class__(
            base_ring=other.base_ring,
            nrows=other.nrows,
            ncols=other.ncols,
            entries=new_entries,
        )

    def __mul__(self, other):
        if (
            str(type(other)) == "<class 'jacamar.rings.reals.RealNumber'>"
            or str(type(other)) == "<class 'jacamar.rings.complexes.ComplexNumber'>"
            or str(type(other)) == "<class 'jacamar.rings.integers.Integer'>"
            or str(type(other)) == "<class 'jacamar.rings.rationals.Rational'>"
        ):
            new_entries = []
            for i in self.entries:
                new_entries.append([])
                for j in i:
                    new_entries[-1].append(other * j)

            return self.__class__(
                base_ring=self.base_ring,
                nrows=self.nrows,
                ncols=self.ncols,
                entries=new_entries,
            )
        if self.ncols != other.nrows:
            raise ValueError(
                f"Cannot multiply matrix of size {self.nrows}x{self.ncols} with matrix of size {other.nrows}x{other.ncols}."
            )

        if self.nrows == 0 or other.nrows == 0:

            return other.__class__(
                base_ring=other.base_ring,
                nrows=self.nrows,
                ncols=other.ncols,
                entries=[],
            )

        # STRASSEN
        elif (
            self.ncols % 2 == 0
            and self.nrows % 2 == 0
            and other.ncols % 2 == 0
            and other.nrows % 2 == 0
            and self.size() == other.size()
            and self.size()[0] > MATRIX_SWITCH
        ):

            def strassen(A, B):
                n = len(A.entries)
                if n <= 2:  # Base case
                    return B.__class__(
                        base_ring=B.base_ring,
                        nrows=n,
                        ncols=n,
                        entries=np.dot(A.entries, B.entries),
                    )

                # Partitions
                mid = n // 2
                A11 = A[:mid, :mid]
                A12 = A[:mid, mid:]
                A21 = A[mid:, :mid]
                A22 = A[mid:, mid:]
                B11 = B[:mid, :mid]
                B12 = B[:mid, mid:]
                B21 = B[mid:, :mid]
                B22 = B[mid:, mid:]
                # Recursions
                P1 = strassen(A11, B12 - B22)
                P2 = strassen(A11 + A12, B22)
                P3 = strassen(A21 + A22, B11)
                P4 = strassen(A22, B21 - B11)
                P5 = strassen(A11 + A22, B11 + B22)
                P6 = strassen(A12 - A22, B21 + B22)
                P7 = strassen(A11 - A21, B11 + B12)

                # Combine results to form C
                C11 = (P5 + P4 - P2 + P6).entries
                C12 = (P1 + P2).entries
                C21 = (P3 + P4).entries
                C22 = (P5 + P1 - P3 - P7).entries

                C1 = np.hstack((np.array(C11), np.array(C12)))
                C2 = np.hstack((np.array(C21), np.array(C22)))
                C = np.vstack((C1, C2)).tolist()

                # Combine quadrants to form C
                return _MatrixGenericData(
                    base_ring=A.base_ring, nrows=len(C), ncols=len(C[0]), entries=C
                )

            # Set up parallelization
            mid = self.nrows // 2
            A11 = self[:mid, :mid]
            A12 = self[:mid, mid:]
            A21 = self[mid:, :mid]
            A22 = self[mid:, mid:]
            B11 = other[:mid, :mid]
            B12 = other[:mid, mid:]
            B21 = other[mid:, :mid]
            B22 = other[mid:, mid:]
            # Recursions (parallel)

            sub_strassen = [
                strassen(A11, B12 - B22),
                strassen(A11 + A12, B22),
                strassen(A21 + A22, B11),
                strassen(A22, B21 - B11),
                strassen(A11 + A22, B11 + B22),
                strassen(A12 - A22, B21 + B22),
                strassen(A11 - A21, B11 + B12),
            ]

            P1 = sub_strassen[0]
            P2 = sub_strassen[1]
            P3 = sub_strassen[2]
            P4 = sub_strassen[3]
            P5 = sub_strassen[4]
            P6 = sub_strassen[5]
            P7 = sub_strassen[6]

            # Combine results to form C
            C11 = (P5 + P4 - P2 + P6).entries
            C12 = (P1 + P2).entries
            C21 = (P3 + P4).entries
            C22 = (P5 + P1 - P3 - P7).entries

            C1 = np.hstack((np.array(C11), np.array(C12)))
            C2 = np.hstack((np.array(C21), np.array(C22)))
            C = np.vstack((C1, C2)).tolist()

            # Combine quadrants to form C
            return _MatrixGenericData(
                base_ring=other.base_ring, nrows=len(C), ncols=len(C[0]), entries=C
            )

        # standard matrix multiplication
        else:
            se = self.entries
            oe = other.entries
            sr = self.nrows
            on = other.ncols
            new_entries = []
            for i in range(sr):
                new_entries.append([])
                for j in range(on):
                    new_entry = se[i][0] * oe[0][j]
                    for k in range(1, self.ncols):
                        new_entry += se[i][k] * oe[k][j]
                    new_entries[i].append(new_entry)

        return other.__class__(
            base_ring=other.base_ring,
            nrows=self.nrows,
            ncols=other.ncols,
            entries=new_entries,
        )

    def __sub__(self, other):
        if self.nrows != other.nrows or self.ncols != other.ncols:
            raise ValueError(
                f"Cannot add matrix of size {self.nrows}x{self.ncols} to matrix of size {other.nrows}x{other.ncols}."
            )

        if self.nrows == 0 or other.nrows == 0:
            return other.__class__(
                base_ring=other.base_ring,
                nrows=other.nrows,
                ncols=other.ncols,
                entries=[],
            )

        new_entries = [self.entries[i].copy() for i in range(self.nrows)]
        for i in range(self.nrows):
            for j in range(self.ncols):
                new_entries[i][j] -= other.entries[i][j]

        return other.__class__(
            base_ring=other.base_ring,
            nrows=other.nrows,
            ncols=other.ncols,
            entries=new_entries,
        )

    def __str__(self):
        return str(self.entries)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.nrows != other.nrows or self.ncols != other.ncols:
            return False
        if self.nrows == 0 or self.ncols == 0:
            return self.base_ring == other.base_ring
        return self.entries == other.entries

    def __call__(self, other):
        """
        RESERVED for eventual evaluation of matrices on vectors.
        """
        raise NotImplementedError

    def __getitem__(self, args):
        """
        If i=args[0] and j=args[1], returns the jth entry of the ith row.
        """
        for idx in args:
            if isinstance(idx, slice):
                # Parse with submatrix.
                r, c = args
                entries = self.entries
                new_entries = entries[r]
                new_data = []

                if not isinstance(new_entries[0], list):
                    new_data = [new_entries[c]]
                    if not isinstance(new_data[0], list):
                        new_data = [new_data]
                else:
                    for i in new_entries:
                        if not isinstance(i[c], list):
                            new_data.append([i[c]])
                        else:
                            new_data.append(i[c])

                ncols = len(new_data[0])
                nrows = len(new_data)
                if nrows == 1 and ncols == 1:
                    return entries[0][0]

                return self.__class__(
                    base_ring=self.base_ring,
                    nrows=nrows,
                    ncols=ncols,
                    entries=new_data,
                )

        return self.entries[args[0]][args[1]]

    def __setitem__(self, args, val):
        """
        If i=args[0] and j=args[1], sets the jth entry of the ith row to be val.
        """
        self.entries[args[0]][args[1]] = val


class _MatrixPythonData:
    def __init__(self, *, base_ring, shape, entries):
        self.base_ring = base_ring
        self.shape = shape
        self.entries = entries


class Matrix:
    """
    Base class for matrices of numbers or polynomials.
    Numerical matrices are built off of flint mat objects
    while generic (polynomial) matrices use the
    _MatrixGenericData format.
    """

    def __init__(
        self,
        *,
        base_ring,
        nrows=None,
        ncols=None,
        entries=None,
        data=None,
    ):

        self.base_ring = base_ring
        if self.base_ring in {ZZ, QQ, RR, CC, RR_py, ZZ_py}:
            self._is_generic = False
            self._is_python = False

            if self.base_ring in {RR_py, ZZ_py}:
                self._is_python = True
        else:
            self._is_generic = True
            self._is_python = False

        # If data is provided, there is a fast constructor.
        if data is not None:
            self.data = data
            if self._is_generic:
                self.nrows = data.nrows
                self.ncols = data.ncols
            else:
                if self._is_python:
                    self.nrows = data.shape[0]
                    self.ncols = data.shape[1]
                else:
                    self.nrows = data.nrows()
                    self.ncols = data.ncols()
        # Else, if the rows or columns are zero, construct an empty matrix of
        # the appropriate size.
        elif nrows == 0 or ncols == 0:
            if nrows is None or ncols is None:
                raise ValueError("Not enough information provided to construct Matrix.")
            self.nrows = nrows
            self.ncols = ncols

            if self.base_ring == ZZ:
                self.data = flint.fmpz_mat(self.nrows, self.ncols)
            elif self.base_ring == QQ:
                self.data = flint.fmpq_mat(self.nrows, self.ncols)
            elif self.base_ring == RR:
                self.data = flint.arb_mat(self.nrows, self.ncols)
            elif self.base_ring == CC:
                self.data = flint.acb_mat(self.nrows, self.ncols)
            elif self.base_ring == ZZ_py:
                self.data = np.ndarray([self.nrows, self.ncols])
            elif self.base_ring == RR_py:
                self.data = np.ndarray([self.nrows, self.ncols])
            else:
                self.data = _MatrixGenericData(
                    base_ring=self.base_ring,
                    nrows=self.nrows,
                    ncols=self.ncols,
                    entries=[],
                )
        else:
            if (nrows is None or ncols is None) and entries is None:
                raise ValueError("Not enough information provided to construct Matrix.")

            if entries is None:
                self.nrows = nrows
                self.ncols = ncols
                new_entries = self._zero_entries()
            else:
                if isinstance(entries, dict):
                    if nrows is None or ncols is None:
                        raise ValueError(
                            "Not enough information to determine the size of the desired matrix."
                        )
                    self.nrows = nrows
                    self.ncols = ncols
                    new_entries = self._zero_entries()
                    # COERCE the given entries into the base ring, or it's data
                    # class in the non-generic case.
                    for t, e in entries.items():
                        if self._is_generic:
                            new_entries[t[0]][t[1]] = base_ring(e)
                        else:
                            new_entries[t[0]][t[1]] = base_ring(e).data

                # Else, assume that it is a list of lists of base_ring elements.
                else:
                    self.nrows = len(entries)
                    if self.nrows == 0:
                        self.ncols = 0
                    else:
                        self.ncols = len(entries[0])
                    # Now, COERCE the given entries into the base_ring.
                    if self._is_generic:
                        new_entries = []
                        for i in range(self.nrows):
                            new_entries.append([])
                            for j in range(self.ncols):
                                new_entries[i].append(self.base_ring(entries[i][j]))
                    # Unless, the base_ring is special, in which case we use
                    # the underlying data to later construct a FLINT matrix.
                    else:
                        new_entries = []
                        for i in range(self.nrows):
                            new_entries.append([])
                            for j in range(self.ncols):
                                # TODO: fix the coercion here. If the input
                                # entries happen to already be NUTHATCH
                                # classes wrapping FLINT classes, then the
                                # following code does not work!
                                new_entries[i].append(
                                    self.base_ring(entries[i][j]).data
                                )

            if self.base_ring == ZZ:
                self.data = flint.fmpz_mat(new_entries)
            elif self.base_ring == QQ:
                self.data = flint.fmpq_mat(new_entries)
            elif self.base_ring == RR:
                self.data = flint.arb_mat(new_entries)
            elif self.base_ring == CC:
                self.data = flint.acb_mat(new_entries)
            elif self.base_ring == ZZ_py:
                self.data = np.array(new_entries)
            elif self.base_ring == RR_py:
                self.data = np.array(new_entries)

            else:
                self.data = _MatrixGenericData(
                    base_ring=self.base_ring,
                    nrows=self.nrows,
                    ncols=self.ncols,
                    entries=new_entries,
                )

    # Constructor helper function.
    def _zero_entries(self):
        entries = []
        for i in range(self.nrows):
            i_row = []
            for j in range(self.ncols):
                if self._is_generic:
                    i_row.append(self.base_ring(0))
                else:
                    i_row.append(self.base_ring(0).data)
            entries.append(i_row)
        return entries

    def det(self):
        """Alias for `determinenant` method."""
        return self.determinant()

    def determinant(self):
        """Determinant method for matrices."""
        if self.nrows != self.ncols:
            raise ValueError("Matrix must be square.")
        if not self._is_generic:
            if self._is_python:
                return self.base_ring(np.linalg.det(self.data))
            return self.base_ring(self.data.det())
        return self.data.det()

    def size(self):
        """Returns size of a matrix as a tuple (rows, cols)."""
        return (self.nrows, self.ncols)

    def T(self):
        """Alias for the transpose() method."""
        return self.transpose()

    def transpose(self):
        """Returns a transposed copy of self."""
        return Matrix(
            base_ring=self.base_ring,
            nrows=self.ncols,
            ncols=self.nrows,
            data=self.data.transpose(),
        )

    def __add__(self, other):
        """Returns self + other with base ring that of other."""
        return other.__class__(
            base_ring=other.base_ring,
            data=self.data + other.data,
        )

    def __sub__(self, other):
        """Returns self - other with base ring that of other."""
        return other.__class__(
            base_ring=other.base_ring,
            data=self.data - other.data,
        )

    def __mul__(self, other):
        if (
            str(type(other)) == "<class 'jacamar.rings.reals.RealNumber'>"
            or str(type(other)) == "<class 'jacamar.rings.complexes.ComplexNumber'>"
            or str(type(other)) == "<class 'jacamar.rings.integers.Integer'>"
            or str(type(other)) == "<class 'jacamar.rings.rationals.Rational'>"
        ):
            if (
                str(self.base_ring)
                == "<class 'jacamar.rings.polynomials.PolynomialRing'>"
            ):
                return self.__class__(
                    base_ring=self.base_ring,
                    nrows=self.nrows,
                    ncols=self.ncols,
                    data=self.data * other,
                )

            return self.__class__(
                base_ring=self.base_ring,
                nrows=self.nrows,
                ncols=self.ncols,
                data=self.data * other.data,
            )

        if self.ncols != other.nrows:
            raise ValueError(
                f"Cannot multiply matrix of size {self.nrows}x{self.ncols} with matrix of size {other.nrows}x{other.ncols}."
            )

        if self.nrows == 0 or other.nrows == 0 or other.ncols == 0:
            return other.__class__(
                base_ring=other.base_ring,
                nrows=self.nrows,
                ncols=other.ncols,
            )

        return other.__class__(
            base_ring=other.base_ring,
            data=self.data * other.data,
        )

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__repr__()

    def __eq__(self, other):
        if self._is_python and other._is_python:
            return self.data.all() == other.data.all()
        return self.data == other.data

    def __call__(self, i, j):
        """
        RESERVED for future use in evaluation on vectors.
        """
        raise NotImplementedError

    def __getitem__(self, args):
        for idx in args:
            if isinstance(idx, slice):
                if not isinstance(args, tuple):
                    return ValueError(
                        "The matrix slice method takes 2 args [rows, columns], but 1 were given."
                    )

                elif len(args) == 2:
                    r, c = args
                    if not self._is_generic:
                        r, c = args
                        entries = self.data.tolist()
                    else:
                        entries = self.data.entries
                    new_entries = entries[r]
                    new_data = []

                    if not isinstance(new_entries[0], list):
                        new_data = [new_entries[c]]
                        if not isinstance(new_data[0], list):
                            new_data = [new_data]
                    else:
                        for i in new_entries:
                            if not isinstance(i[c], list):
                                new_data.append([i[c]])
                            else:
                                new_data.append(i[c])

                    ncols = len(new_data[0])
                    nrows = len(new_data)
                    if not self._is_generic:
                        r, c = args
                        entries = self.data.tolist()
                        return self.__class__(
                            base_ring=self.base_ring,
                            nrows=nrows,
                            ncols=ncols,
                            entries=new_data,
                        )

                    return self.__class__(
                        base_ring=self.base_ring,
                        nrows=nrows,
                        ncols=ncols,
                        entries=new_data,
                        data=_MatrixGenericData(
                            base_ring=self.base_ring,
                            nrows=nrows,
                            ncols=ncols,
                            entries=new_data,
                        ),
                    )

        return self.base_ring(self.data[args])

    def __setitem__(self, args, val):
        self.data[args] = val


# Functions for matrices


def generate(value, nrows, ncols):
    """
    Generates a repeating matrix of size (rows, cols). Can be generic or not.
    """
    entries = []
    for i in range(nrows):
        entries.append([])
        for j in range(ncols):
            entries[-1].append(value)

    return Matrix(
        base_ring=value.ring,
        nrows=nrows,
        ncols=ncols,
        entries=entries,
        data=_MatrixGenericData(
            base_ring=value.ring,
            nrows=nrows,
            ncols=ncols,
            entries=entries,
        ),
    )


def random(base_ring, max_val, nrows, ncols, poly=0, poly_ring=0, ngens=1):
    """
    Generates a random matrix of size nrows x ncols using np.random.
    Values go from [0, max_val].
    Polynomials are of form n0*x0**r0 + n1*x1**r1 ....
    """
    entries = []

    if poly:
        p = poly_ring(base_ring=base_ring, ngens=ngens, prefix="x", packed=True)
        for i in range(nrows):
            entries.append([])
            for j in range(ncols):
                val = p(0)
                for k in range(ngens):
                    if base_ring == RR:
                        val = val + base_ring(np.random.random() * max_val) * p.gens[
                            k
                        ] ** ZZ(int(np.random.random() * max_val))
                    elif base_ring == ZZ:
                        val = val + base_ring(
                            int(np.random.random() * max_val)
                        ) * p.gens[k] ** ZZ(int(np.random.random() * max_val))

                entries[-1].append(val)

        return Matrix(
            base_ring=poly_ring,
            nrows=nrows,
            ncols=ncols,
            entries=entries,
            data=_MatrixGenericData(
                base_ring=poly_ring,
                nrows=nrows,
                ncols=ncols,
                entries=entries,
            ),
        )
    if base_ring == RR:
        vals = np.random.random(nrows * ncols) * max_val
        for i in range(nrows):
            entries.append([])
            for j in range(ncols):
                entries[-1].append(base_ring(vals[(i + 1) * (j)]))

    elif base_ring == ZZ:
        vals = np.random.randint(max_val, size=nrows * ncols)
        for i in range(nrows):
            entries.append([])
            for j in range(ncols):
                entries[-1].append(base_ring(int(vals[(i + 1) * (j)])))

    elif base_ring == RR_py:
        vals = np.random.random(nrows * ncols) * max_val
        for i in range(nrows):
            entries.append([])
            for j in range(ncols):
                entries[-1].append(vals[(i + 1) * (j)])

        return Matrix(
            base_ring=base_ring,
            entries=entries,
        )
    elif base_ring == ZZ_py:
        vals = np.random.randint(max_val, size=nrows * ncols)
        for i in range(nrows):
            entries.append([])
            for j in range(ncols):
                entries[-1].append(int(vals[(i + 1) * (j)]))

        return Matrix(
            base_ring=base_ring,
            entries=entries,
        )

    return Matrix(
        base_ring=base_ring,
        nrows=nrows,
        ncols=ncols,
        entries=entries,
    )



"""
C-MAT
Cython implementation of matrix multiplication methods. 
Compiled in C in the file cmat.c
"""
import cython
def c_generic_mul(se, oe, sr, on):
    new_entries = []
    for i in range(sr):
        new_entries.append([])
        for j in range(on):
            new_entry = se[i][0] * oe[0][j]
            for k in range(1, self.ncols):
                new_entry += se[i][k] * oe[k][j]
            new_entries[i].append(new_entry)
    return new_entries

